import copy
import datetime
import hashlib
import json
import logging
import os
import re
import shlex
import subprocess
from asyncio import Event

from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from runhouse.constants import (
    CLUSTER_CONFIG_PATH,
    LAST_ACTIVE_AT_TIMEFRAME,
    RESERVED_SYSTEM_NAMES,
    SKY_VENV,
    TIME_UNITS,
)
from runhouse.globals import configs, rns_client

from runhouse.logger import get_logger
from runhouse.resources.hardware.sky.command_runner import (
    _HASH_MAX_LENGTH,
    ssh_options_list,
    SshMode,
)
from runhouse.utils import ColoredFormatter, run_setup_command

logger = get_logger(__name__)


class ServerConnectionType(str, Enum):
    """Manage the type of connection Runhouse will make with the API server started on the cluster.
    ``ssh``: Use port forwarding to connect to the server via SSH, by default on port 32300.
    ``tls``: Do not use port forwarding and start the server with HTTPS (using custom or fresh TLS certs), by default
        on port 443.
    ``none``: Do not use port forwarding, and start the server with HTTP, by default on port 80.
    """

    SSH = "ssh"
    TLS = "tls"
    NONE = "none"


class LauncherType(str, Enum):
    LOCAL = "local"
    DEN = "den"


class RunhouseDaemonStatus(str, Enum):
    RUNNING = "running"
    TERMINATED = "terminated"
    UNAUTHORIZED = "unauthorized"
    UNKNOWN = "unknown"
    INTERNAL_SERVER_ERROR = "internal_server_error"
    RUNHOUSE_DAEMON_DOWN = "runhouse_daemon_down"
    INVALID_URL = "invalid_url"


class ClusterStatus(str, Enum):
    RUNNING = "running"
    TERMINATED = "terminated"
    INITIALIZING = "initializing"
    UNKNOWN = "unknown"


def cluster_config_file_exists() -> bool:
    return Path(CLUSTER_CONFIG_PATH).expanduser().exists()


def load_cluster_config_from_file() -> Dict:
    if cluster_config_file_exists():
        with open(Path(CLUSTER_CONFIG_PATH).expanduser()) as f:
            cluster_config = json.load(f)
        return cluster_config
    else:
        return {}


def _config_and_args_mismatches(config, alt_options):
    """Overload by child resources to compare their config with the alt_options. If the user specifies alternate
    options, compare the config with the options. It's generally up to the child class to decide how to handle the
    options, but default behavior is provided. The default behavior simply checks if any of the alt_options are
    present in the config (with awareness of resources), and if their values differ, return None.

    If the child class returns None, it's deciding to override the config
    with the options. If the child class returns a config, it's deciding to use the config and ignore the options
    (or somehow incorporate them, rarely). Note that if alt_options are provided and the config is not found,
    no error is raised, while if alt_options are not provided and the config is not found, an error is raised.
    """
    from runhouse.resources.images.image import Image
    from runhouse.resources.resource import Resource

    def alt_option_to_repr(val):
        if isinstance(val, dict):
            # This can either be a sub-resource which hasn't been converted to a resource yet, or an
            # actual user-provided dict
            if "rns_address" in val:
                return val["rns_address"]
            if "name" in val:
                # convert a user-provided name to an rns_address
                return rns_client.resolve_rns_path(val["name"])
            else:
                return val
        elif isinstance(val, list):
            val = [str(item) if isinstance(item, int) else item for item in val]
        elif isinstance(val, int) or isinstance(val, float):
            val = str(val)
        elif isinstance(val, Image):
            val = val.config()
        elif isinstance(val, Resource) and (
            val.config().get("name") or val.config().get("rns_address")
        ):
            return alt_option_to_repr(val.config())
        return val

    mismatches = {}
    for key, value in alt_options.items():
        if key in config:
            if alt_option_to_repr(value) != alt_option_to_repr(config[key]):
                mismatches[key] = value
        else:
            mismatches[key] = value
    return mismatches


def _current_cluster(key="config"):
    """Retrieve key value from the current cluster config.
    If key is "config", returns entire config."""
    from runhouse.globals import obj_store

    cluster_config = obj_store.get_cluster_config()
    cluster_config.pop("creds", None)
    if cluster_config:
        # This could be a local cluster started via runhouse server start,
        # in which case it would have no Name.
        if key in ["cluster_name", "name"] and "name" not in cluster_config:
            return None
        if key == "config":
            return cluster_config
        if key == "cluster_name":
            return cluster_config["name"].rsplit("/", 1)[-1]
        return cluster_config[key]
    else:
        return None


def _get_cluster_from(system, dryrun=False):
    from .cluster import Cluster

    if isinstance(system, Cluster):
        return system
    if system in RESERVED_SYSTEM_NAMES:
        return system

    if isinstance(system, Dict):
        return Cluster.from_config(system, dryrun)

    if isinstance(system, str):
        config = _current_cluster(key="config")
        if config and system == config.get("name"):
            return Cluster.from_config(config, dryrun)
        try:
            system = Cluster.from_name(name=system, dryrun=dryrun)
        except ValueError:
            # Name not found in Den. Doing the lookup this way saves us a hop to Den
            pass

    return system


def _setup_default_creds(cluster_type: str):
    from runhouse.resources.secrets import Secret

    default_ssh_key = rns_client.default_ssh_key
    if cluster_type == "OnDemandCluster":
        try:
            sky_secret = Secret.from_name("sky")
            return sky_secret
        except ValueError:
            if default_ssh_key:
                # copy over default key to sky-key for launching use
                default_secret = Secret.from_name(default_ssh_key)
                sky_secret = default_secret._write_to_file("~/.ssh/sky-key")
                return sky_secret
            else:
                return None
    elif default_ssh_key:
        return Secret.from_name(default_ssh_key)
    return None


def _setup_creds_from_dict(ssh_creds: Dict, cluster_name: str):
    from runhouse.resources.secrets import Secret
    from runhouse.resources.secrets.provider_secrets.sky_secret import SkySecret
    from runhouse.resources.secrets.provider_secrets.ssh_secret import SSHSecret

    creds = copy.copy(ssh_creds)
    ssh_properties = {}
    cluster_secret = None

    private_key_path = creds["ssh_private_key"] if "ssh_private_key" in creds else None
    password = creds.pop("password") if "password" in creds else None

    if private_key_path:
        key = os.path.basename(private_key_path)

        if password:
            # extract ssh values and create ssh secret
            values = (
                SSHSecret.extract_secrets_from_path(private_key_path)
                if private_key_path
                else {}
            )
            values["password"] = password
            cluster_secret = SSHSecret(
                name=f"{cluster_name}-ssh-secret",
                provider="ssh",
                key=key,
                path=private_key_path,
                values=values,
            )
        else:
            # set as standard SSH secret
            constructor = SkySecret if key == "sky-key" else SSHSecret
            cluster_secret = constructor(
                name=f"ssh-{key}", key=key, path=private_key_path
            )
    elif password:
        cluster_secret = Secret(
            name=f"{cluster_name}-ssh-secret", values={"password": password}
        )

    # keep track of non secret/password values in ssh_properties
    if isinstance(creds, Dict):
        ssh_properties = creds

    return cluster_secret, ssh_properties


def is_gpu_cluster(cluster: "Cluster" = None, node: Optional[str] = None):
    if run_setup_command("nvidia-smi", cluster=cluster, node=node)[0] == 0:
        return True
    return False


def detect_cuda_version_or_cpu(cluster: "Cluster" = None, node: Optional[str] = None):
    """Return the CUDA version on the cluster. If we are on a CPU-only cluster return 'cpu'.

    Note: A cpu-only machine may have the CUDA toolkit installed, which means nvcc will still return
    a valid version. Also check if the NVIDIA driver is installed to confirm we are on a GPU."""

    if not is_gpu_cluster(cluster=cluster, node=node):
        return "cpu"

    status_codes = run_setup_command("nvcc --version", cluster=cluster, node=node)
    if not status_codes[0] == 0:
        status_codes = run_setup_command(
            "/usr/local/cuda/bin/nvcc --version", cluster=cluster, node=node
        )
        if not status_codes[0] == 0:
            raise RuntimeError(
                "Could not determine CUDA version on GPU cluster for installing the correct torch version. "
                "Please install nvcc on the cluster to enable automatic CUDA version detection, or include "
                "the exact version to install for the package, e.g. 'torch==1.13.1+cu117' or by including "
                "the --index-url or --extra-index-url"
            )
    cuda_version = status_codes[1].split("release ")[1].split(",")[0]

    return cuda_version


def _run_ssh_command(
    address: str,
    ssh_user: str,
    ssh_port: int,
    ssh_private_key: str,
    ssh_proxy_command: str,
    docker_user: str,
):
    from runhouse.resources.hardware.sky_command_runner import SkySSHRunner

    runner = SkySSHRunner(
        (address, ssh_port),
        ssh_user=ssh_user,
        ssh_private_key=ssh_private_key,
        ssh_proxy_command=ssh_proxy_command,
        docker_user=docker_user,
    )
    ssh_command = runner._ssh_base_command(
        ssh_mode=SshMode.INTERACTIVE, port_forward=None
    )
    subprocess.run(ssh_command)


def _docker_ssh_proxy_command(
    address: str,
    ssh_user: str,
    ssh_private_key: str,
):
    return lambda ssh: " ".join(
        ssh
        + ssh_options_list(ssh_private_key, None)
        + ["-W", "%h:%p", f"{ssh_user}@{address}"]
    )


# Adapted from SkyPilot Command Runner
def _ssh_base_command(
    address: str,
    ssh_user: str,
    ssh_private_key: str,
    ssh_control_name: Optional[str] = "__default__",
    ssh_proxy_command: Optional[str] = None,
    ssh_port: int = 22,
    docker_ssh_proxy_command: Optional[str] = None,
    disable_control_master: Optional[bool] = False,
    ssh_mode: SshMode = SshMode.INTERACTIVE,
    port_forward: Optional[List[int]] = None,
    connect_timeout: Optional[int] = 30,
):
    ssh = ["ssh"]
    if ssh_mode == SshMode.NON_INTERACTIVE:
        # Disable pseudo-terminal allocation. Otherwise, the output of
        # ssh will be corrupted by the user's input.
        ssh += ["-T"]
    else:
        # Force pseudo-terminal allocation for interactive/login mode.
        ssh += ["-tt"]

    if port_forward is not None:
        # RH MODIFIED: Accept port int (to forward same port) or pair of ports
        for fwd in port_forward:
            if isinstance(fwd, int):
                local, remote = fwd, fwd
            else:
                local, remote = fwd
            logger.debug(f"Forwarding port {local} to port {remote} on localhost.")
            ssh += ["-L", f"{local}:localhost:{remote}"]

    return (
        ssh
        + ssh_options_list(
            ssh_private_key,
            ssh_control_name,
            ssh_proxy_command=ssh_proxy_command,
            docker_ssh_proxy_command=docker_ssh_proxy_command,
            # TODO change to None like before?
            port=ssh_port,
            connect_timeout=connect_timeout,
            disable_control_master=disable_control_master,
        )
        + [f"{ssh_user}@{address}"]
    )


def _generate_ssh_control_hash(ssh_control_name):
    return hashlib.md5(ssh_control_name.encode()).hexdigest()[:_HASH_MAX_LENGTH]


def _cluster_set_autostop_command(autostop_mins: int):
    sky_set_autostop_cmd = shlex.quote(
        f"from sky.skylet.autostop_lib import set_autostop; "
        f'set_autostop({autostop_mins}, "cloudvmray", True)'
    )
    return f"{SKY_VENV}/bin/python -c {sky_set_autostop_cmd}"


def up_cluster_helper(cluster, capture_output: Union[bool, str] = True):
    from runhouse.utils import SuppressStd

    if capture_output:
        try:
            with SuppressStd() as outfile:
                cluster.up()
        except Exception as e:
            if isinstance(capture_output, str):
                logger.error(
                    f"Error starting cluster {cluster.name}, logs written to {capture_output}"
                )
            raise e
        finally:
            if isinstance(capture_output, str):
                with open(capture_output, "w") as f:
                    f.write(outfile.output)
    else:
        cluster.up()


###################################
# Cluster list helping methods
###################################
def parse_time_duration(duration: str):
    # A simple parser for duration like "15m", "2h", "3d"
    try:
        unit = duration[-1]
        value = int(duration[:-1])

        time_filter_match = re.match(r"(\d+)([smhd])$", duration)

        if time_filter_match:

            # if the user provides a "--since" time filter than is less than a minute, set it by default to one minute.
            if unit == "s" and value < 60:
                value = 60

            time_duration = value * TIME_UNITS.get(unit)

        else:
            logger.warning(
                f"Filter is not applied, invalid time unit provided ({unit}). Setting filter to default (24h)."
            )
            time_duration = LAST_ACTIVE_AT_TIMEFRAME

        return time_duration
    except Exception:
        return LAST_ACTIVE_AT_TIMEFRAME


def parse_filters(since: str, cluster_status: Union[str, ClusterStatus]):
    cluster_filters = {}

    if since:
        last_active_in: int = parse_time_duration(
            duration=since
        )  # return in representing the "since" filter in seconds
        if last_active_in:
            cluster_filters["since"] = last_active_in

    if cluster_status:
        cluster_filters["cluster_status"] = cluster_status

    return cluster_filters


def get_clusters_from_den(cluster_filters: dict, force: bool):
    get_clusters_params = {"resource_type": "cluster", "owned": True}

    if (
        "cluster_status" in cluster_filters
        and cluster_filters["cluster_status"] == ClusterStatus.TERMINATED
    ):
        # Include the relevant daemon status for the filter
        cluster_filters["daemon_status"] = RunhouseDaemonStatus.TERMINATED

    # If "all" filter is specified load all clusters (no filters are added to get_clusters_params)
    if cluster_filters and "all" not in cluster_filters.keys():
        get_clusters_params.update(cluster_filters)

    # If not filters are specified, get only running clusters.
    elif not cluster_filters:
        # For ondemand clusters, use cluster status (via Sky). for other clusters, use the status of the daemon
        get_clusters_params.update(
            {
                "cluster_status": ClusterStatus.RUNNING,
                "daemon_status": RunhouseDaemonStatus.RUNNING,
                "since": LAST_ACTIVE_AT_TIMEFRAME,
                "force": force,
            }
        )

    clusters_in_den_resp = rns_client.session.get(
        f"{rns_client.api_server_url}/resource",
        params=get_clusters_params,
        headers=rns_client.request_headers(),
    )

    return clusters_in_den_resp


def get_unsaved_live_clusters(den_clusters: List[Dict]):
    import sky

    try:
        sky_live_clusters: list = sky.status()
        den_clusters_names = [c.get("name") for c in den_clusters]

        # getting the on-demand clusters that are not saved in den.
        if sky_live_clusters:
            return [
                cluster
                for cluster in sky_live_clusters
                if f'/{rns_client.username}/{cluster.get("name")}'
                not in den_clusters_names
            ]
        else:
            return []
    except Exception as e:
        logger.debug(f"Failed to get unsaved sky live clusters: {e}")
        return []


def get_all_sky_clusters():
    import sky

    try:
        sky_live_clusters: list = sky.status()

        # getting the on-demand clusters that are not saved in den.
        if sky_live_clusters:
            return [cluster.get("name") for cluster in sky_live_clusters]
        else:
            return []
    except Exception as e:
        logger.debug(f"Failed to get sky live clusters: {e}")
        return []


def cast_last_active_timestamp(clusters: List[Dict[str, Any]]):
    for cluster in clusters:
        timestamp = cluster.get("Last Active (UTC)")
        timestamp = (
            timestamp.strftime("%m/%d/%Y, %H:%M:%S")
            if isinstance(timestamp, datetime.datetime)
            else timestamp
        )
        timestamp = timestamp if timestamp != "01/01/1970, 00:00:00" else None
        cluster["Last Active (UTC)"] = timestamp
    return clusters


def get_running_and_not_running_clusters(clusters: list):
    up_clusters, down_clusters = [], []
    for den_cluster in clusters:
        cluster_name = den_cluster.get("name")
        cluster_type = den_cluster.get("data").get("resource_subtype")
        cluster_status = den_cluster.get("cluster_status")
        cluster_status = cluster_status or ClusterStatus.UNKNOWN.value

        # The split is required to remove milliseconds and the offset (according to UTC) from the timestamp.
        # (cluster_status_last_checked is in the following format: YYYY-MM-DD HH:MM:SS.ssssss±HH:MM)
        last_active_at = den_cluster.get("cluster_status_last_checked")

        # Convert to datetime
        last_active_at = (
            datetime.datetime.fromisoformat(last_active_at.split(".")[0])
            if isinstance(last_active_at, str)
            else datetime.datetime(1970, 1, 1)
        )

        last_active_at = last_active_at.replace(tzinfo=datetime.timezone.utc)
        if cluster_status == ClusterStatus.RUNNING and not last_active_at:
            # For BC, in case there are clusters that were saved and created before sending cluster status to Den
            cluster_status = ClusterStatus.UNKNOWN.value

        cluster_info = {
            "Name": cluster_name,
            "Cluster Type": cluster_type,
            "Status": cluster_status,
            "Autostop": den_cluster.get("data", {}).get("autostop_mins"),
            "Last Active (UTC)": last_active_at,
        }

        if cluster_status == ClusterStatus.RUNNING:
            up_clusters.append(cluster_info)
        else:
            down_clusters.append(cluster_info)

    # Sort clusters by the 'Last Active (UTC)' and 'Status' column
    down_clusters = sorted(
        down_clusters,
        key=lambda x: (x["Last Active (UTC)"], x["Status"]),
        reverse=True,
    )

    down_clusters = cast_last_active_timestamp(clusters=down_clusters)
    up_clusters = sorted(
        up_clusters, key=lambda x: x["Last Active (UTC)"], reverse=True
    )

    return cast_last_active_timestamp(clusters=up_clusters), cast_last_active_timestamp(
        clusters=down_clusters
    )


def pprint_launched_cluster_summary(cluster):
    properties: dict = copy.deepcopy(cluster.compute_properties)
    cluster_name = cluster.name
    rns_address = cluster.rns_address
    properties["autostop_mins"] = cluster.autostop_mins

    blue = ColoredFormatter.get_color("blue")
    white = ColoredFormatter.get_color("white")
    cyan = ColoredFormatter.get_color("cyan")
    green = ColoredFormatter.get_color("green")
    reset = ColoredFormatter.get_color("reset")
    italic = ColoredFormatter.get_color("italic")

    if rns_address:
        print(
            f"► To view the cluster in Den, visit https://run.house/resources/{rns_client.format_rns_address(rns_address)}"
        )
    print(f"{blue}Launch Summary:{reset}")
    for key, value in properties.items():
        if value is None:
            continue
        bullet = f"{white}•{reset}"
        key_str = f"{blue}{key.replace('_', ' ').capitalize()}{reset}"
        value_str = f"{green}{value}{reset}"
        print(f"{bullet} {key_str}: {value_str}")

    print(
        f"To teardown the cluster, run {italic}{cyan}runhouse cluster down {cluster_name}{reset}, "
        f'or in python via {italic}{cyan}rh.cluster("{cluster_name}").teardown(){reset}'
    )


###################################
# CLUSTER LOGS
###################################
def get_saved_logs_from_den(rns_address: str):
    """
    get the latest cluster logs saved in den.
    """
    cluster_uri = rns_client.resource_uri(rns_address)
    clusters_in_den_resp = rns_client.session.get(
        f"{rns_client.api_server_url}/resource/{cluster_uri}/logs",
        headers=rns_client.request_headers(),
    )
    return clusters_in_den_resp


async def stream_logs_from_url(
    stop_event: Event, url: str, temp_dir: str, launch_id: str
):
    """Load logs returned from the specified Den URL. The response should be a stream of JSON logs."""
    import httpx

    client = httpx.AsyncClient(timeout=None)

    async with client.stream(
        "POST",
        url,
        json={"temp_dir": temp_dir, "launch_id": launch_id},
        headers=rns_client.request_headers(),
    ) as res:
        if res.status_code != 200:
            error_resp = await res.aread()
            raise ValueError(f"Error calling Den logs API: {error_resp.decode()}")

        async for response_json in res.aiter_lines():
            if stop_event.is_set():
                break
            resp = json.loads(response_json)

            # TODO [JL] any formatting to do here?
            print(resp)

    await client.aclose()


def load_logs_in_thread(stop_event: Event, url: str, temp_dir: str, launch_id: str):
    import asyncio

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(stream_logs_from_url(stop_event, url, temp_dir, launch_id))


###################################
# SERVER-SIDE EVENTS
###################################
_FIELD_SEPARATOR = ":"


class SSEClient(object):
    """Implementation of a SSE client.

    See http://www.w3.org/TR/2009/WD-eventsource-20091029/ for the
    specification.
    """

    def __init__(self, event_source, char_enc="utf-8"):
        """Initialize the SSE client over an existing, ready to consume
        event source.

        The event source is expected to be a binary stream and have a close()
        method. That would usually be something that implements
        io.BinaryIOBase, like an httplib or urllib3 HTTPResponse object.
        """
        self._logger = logging.getLogger(self.__class__.__module__)
        self._logger.debug("Initialized SSE client from event source %s", event_source)
        self._event_source = event_source
        self._char_enc = char_enc

    def _read(self):
        """Read the incoming event source stream and yield event chunks.

        Unfortunately it is possible for some servers to decide to break an
        event into multiple HTTP chunks in the response. It is thus necessary
        to correctly stitch together consecutive response chunks and find the
        SSE delimiter (empty new line) to yield full, correct event chunks."""
        data = b""
        for chunk in self._event_source:
            for line in chunk.splitlines(True):
                data += line
                if data.endswith((b"\r\r", b"\n\n", b"\r\n\r\n")):
                    yield data
                    data = b""
        if data:
            yield data

    def events(self):
        for chunk in self._read():
            event = Event()
            # Split before decoding so splitlines() only uses \r and \n
            for line in chunk.splitlines():
                # Decode the line.
                line = line.decode(self._char_enc)

                # Lines starting with a separator are comments and are to be
                # ignored.
                if not line.strip() or line.startswith(_FIELD_SEPARATOR):
                    continue

                data = line.split(_FIELD_SEPARATOR, 1)
                field = data[0]

                # Ignore unknown fields.
                if field not in event.__dict__:
                    self._logger.debug(
                        "Saw invalid field %s while parsing " "Server Side Event", field
                    )
                    continue

                if len(data) > 1:
                    # From the spec:
                    # "If value starts with a single U+0020 SPACE character,
                    # remove it from value."
                    if data[1].startswith(" "):
                        value = data[1][1:]
                    else:
                        value = data[1]
                else:
                    # If no value is present after the separator,
                    # assume an empty value.
                    value = ""

                # The data field may come over multiple lines and their values
                # are concatenated with each other.
                if field == "data":
                    event.__dict__[field] += value + "\n"
                else:
                    event.__dict__[field] = value

            # Events with no data are not dispatched.
            if not event.data:
                continue

            # If the data field ends with a newline, remove it.
            if event.data.endswith("\n"):
                event.data = event.data[0:-1]

            # Empty event names default to 'message'
            event.event = event.event or "message"

            # Dispatch the event
            self._logger.debug("Dispatching %s...", event)
            yield event

    def close(self):
        """Manually close the event source stream."""
        self._event_source.close()


class Event(object):
    """Representation of an event from the event stream."""

    def __init__(self, id=None, event="message", data="", retry=None):
        self.id = id
        self.event = event
        self.data = data
        self.retry = retry

    def __str__(self):
        s = "{0} event".format(self.event)
        if self.id:
            s += " #{0}".format(self.id)
        if self.data:
            s += ", {0} byte{1}".format(len(self.data), "s" if len(self.data) else "")
        else:
            s += ", no data"
        if self.retry:
            s += ", retry in {0}ms".format(self.retry)
        return s


###################################
# KUBERNETES SETUP
###################################
def setup_kubernetes(
    kube_namespace: Optional[str] = None,
    kube_config_path: Optional[str] = None,
    kube_context: Optional[str] = None,
    **kwargs,
):
    if kwargs.get("provider") and not kwargs.get("provider") == "kubernetes":
        raise ValueError(
            f"Received non kubernetes provider {kwargs.get('provider')} with kubernetes specific "
            "cluster arguments."
        )

    if (
        kwargs.get("server_connection_type")
        and kwargs.get("server_connection_type") != ServerConnectionType.SSH
    ):
        raise ValueError(
            "Runhouse K8s Cluster server connection type must be set to `ssh`. "
            f"You passed {kwargs.get('server_connection_type')}."
        )

    if kube_context and kube_namespace:
        logger.warning(
            "You passed both a context and a namespace. Ensure your namespace matches the one in your context.",
        )

    launcher = kwargs.get("launcher") or configs.launcher
    if launcher == "local":
        if kube_context:
            # check if user passed a user-defined context
            try:
                cmd = f"kubectl config use-context {kube_context}"  # set user-defined context as current context
                subprocess.run(cmd, shell=True, check=True)
                logger.info(f"Kubernetes context has been set to: {kube_context}")
            except subprocess.CalledProcessError as e:
                logger.error(f"Error setting context {kube_context}: {e}")

        if kube_namespace:
            # Set the context only if launching locally
            # check if user passed a user-defined namespace
            cmd = f"kubectl config set-context --current --namespace={kube_namespace}"
            try:
                process = subprocess.run(
                    cmd,
                    shell=True,
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                )
                logger.debug(process.stdout)
                logger.info(f"Kubernetes namespace set to {kube_namespace}")

            except subprocess.CalledProcessError as e:
                logger.info(f"Error: {e}")

        if kube_config_path:  # check if user passed a user-defined kube_config_path
            kube_config_dir = os.path.expanduser("~/.kube")
            kube_config_path_rl = os.path.join(kube_config_dir, "config")

            if not os.path.exists(
                kube_config_dir
            ):  # check if ~/.kube directory exists on local machine
                try:
                    os.makedirs(
                        kube_config_dir
                    )  # create ~/.kube directory if it doesn't exist
                    logger.info(f"Created directory: {kube_config_dir}")
                except OSError as e:
                    logger.info(f"Error creating directory: {e}")

            if os.path.exists(kube_config_path_rl):
                raise Exception(
                    "A kubeconfig file already exists in ~/.kube directory. Aborting."
                )

            try:
                cmd = f"cp {kube_config_path} {kube_config_path_rl}"  # copy user-defined kube_config to ~/.kube/config
                subprocess.run(cmd, shell=True, check=True)
                logger.info(f"Copied kubeconfig to: {kube_config_path}")
            except subprocess.CalledProcessError as e:
                logger.info(f"Error copying kubeconfig: {e}")



import asyncio
import copy
import datetime
import json
import os
import threading
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import httpx

import requests

import runhouse

from runhouse.constants import (
    DEFAULT_AUTOSTOP_CHECK_INTERVAL,
    DEFAULT_LOG_LEVEL,
    DEFAULT_STATUS_CHECK_INTERVAL,
    GPU_COLLECTION_INTERVAL,
    INCREASED_GPU_COLLECTION_INTERVAL,
    INCREASED_STATUS_CHECK_INTERVAL,
    MAX_GPU_INFO_LEN,
    REDUCED_GPU_INFO_LEN,
    SERVER_LOGFILE,
    SERVER_LOGS_FILE_NAME,
)

from runhouse.globals import configs, obj_store, rns_client
from runhouse.logger import get_logger
from runhouse.resources.hardware import load_cluster_config_from_file
from runhouse.resources.hardware.ray_utils import kill_actors
from runhouse.resources.hardware.utils import is_gpu_cluster
from runhouse.rns.rns_client import ResourceStatusData
from runhouse.rns.utils.api import ResourceAccess
from runhouse.servers.autostop_helper import AutostopHelper
from runhouse.servers.http.auth import AuthCache
from runhouse.servers.http.http_utils import CreateProcessParams
from runhouse.utils import (
    ColoredFormatter,
    get_gpu_usage,
    get_pid,
    ServletType,
    sync_function,
)

logger = get_logger(__name__)


class ClusterServletError(Exception):
    pass


class ClusterServlet:
    async def __init__(
        self, cluster_config: Optional[Dict[str, Any]] = None, *args, **kwargs
    ):
        import psutil

        # We do this here instead of at the start of the HTTP Server startup
        # because someone can be running `HTTPServer()` standalone in a test
        # and still want an initialized cluster config in the servlet.
        if not cluster_config:
            cluster_config = load_cluster_config_from_file()
        self._cluster_config: Optional[Dict[str, Any]] = cluster_config

        self._initialized_servlet_args: Dict[str, CreateProcessParams] = {}
        self._key_to_servlet_name: Dict[Any, str] = {}
        self._auth_cache: AuthCache = AuthCache()
        self._paths_to_prepend_in_new_processes = []
        self._env_vars_to_set_in_new_processes = {}
        self._node_servlet_names: List[str] = []
        self._cluster_name = self._cluster_config.get("name", None)
        self._cluster_uri = (
            rns_client.format_rns_address(self._cluster_name)
            if self._cluster_name
            else None
        )

        # will be used in the periodic loop which sends cluster logs to den. Since we are saving logs to the server.log
        # file asynchronously, there might be a use case where we want to send logs to den but the server.log has not
        # been created yet, since it is a freshly initialized cluster servlet. In this case, we will return an empty string.
        self._is_log_file_ready = False

        # will need this to make sure that we are deleting the correct cluster servlet when we call kill_actors
        self.name = kwargs.get("name", "cluster_servlet")

        self.autostop_helper = None
        if self._cluster_config.get("resource_subtype", None) == "OnDemandCluster":
            self.autostop_helper = AutostopHelper()

        self.pid = get_pid()
        self.process = psutil.Process(pid=self.pid)
        self.gpu_metrics = None  # will be updated only if this is a gpu cluster.
        # will be used when self.gpu_metrics will be updated by different threads.
        self.lock = threading.Lock()

        self._cluster_config["is_gpu"] = is_gpu_cluster()

        # creating periodic check loops
        if self._cluster_config.get("is_gpu"):
            logger.debug("Creating _periodic_gpu_check thread.")
            collect_gpu_thread = threading.Thread(
                target=self._periodic_gpu_check, daemon=True
            )
            collect_gpu_thread.start()

        logger.debug("Creating periodic_cluster_checks thread.")
        self.cluster_checks_thread = threading.Thread(
            target=self.periodic_cluster_checks, daemon=True
        )
        self.cluster_checks_thread.start()

        logger.debug("Creating autostop check thread.")
        self.autostop_check_thread = threading.Thread(
            target=self.periodic_autostop_check, daemon=True
        )
        self.autostop_check_thread.start()

    ##############################################
    # List of node servlet names
    ##############################################
    async def aget_node_servlet_names(self) -> List[str]:
        return self._node_servlet_names

    async def aset_node_servlet_names(self, node_servlet_names: List[str]):
        self._node_servlet_names = node_servlet_names

    ##############################################
    # Add to path and get path methods
    ##############################################
    async def aadd_path_to_prepend_in_new_processes(self, path: str):
        self._paths_to_prepend_in_new_processes.append(path)

    async def aget_paths_to_prepend_in_new_processes(self) -> List[str]:
        return self._paths_to_prepend_in_new_processes

    ##############################################
    # Env vars to set and get env vars methods
    ##############################################
    async def aset_env_vars_globally(self, env_vars: Dict[str, Any]):

        await asyncio.gather(
            *[
                obj_store.acall_servlet_method(
                    servlet_name,
                    "aset_env_vars",
                    env_vars,
                    use_servlet_cache=False,
                )
                for servlet_name in await self.aget_all_initialized_servlet_args()
            ]
        )

        await self.aadd_env_vars_to_set_in_new_processes(env_vars)

    async def aadd_env_vars_to_set_in_new_processes(self, env_vars: Dict[str, Any]):
        self._env_vars_to_set_in_new_processes.update(env_vars)

    async def aget_env_vars_to_set_in_new_processes(self) -> Dict[str, str]:
        return self._env_vars_to_set_in_new_processes

    ##############################################
    # Cluster config state storage methods
    ##############################################
    async def aget_cluster_config(self) -> Dict[str, Any]:
        return self._cluster_config

    async def aset_cluster_config(self, cluster_config: Dict[str, Any]):
        if "is_gpu" not in cluster_config.keys():
            cluster_config["is_gpu"] = is_gpu_cluster()

        self._cluster_config = cluster_config

        # Propagate the changes to all other process's obj_stores
        await asyncio.gather(
            *[
                obj_store.acall_servlet_method(
                    servlet_name,
                    "aset_cluster_config",
                    cluster_config,
                    use_servlet_cache=False,
                )
                for servlet_name in await self.aget_all_initialized_servlet_args()
            ]
        )

        return self._cluster_config

    async def aset_cluster_config_value(self, key: str, value: Any):
        if self.autostop_helper and key == "autostop_mins":
            await self.autostop_helper.set_autostop(value)
        self._cluster_config[key] = value

        # Propagate the changes to all other process's obj_stores
        await asyncio.gather(
            *[
                obj_store.acall_servlet_method(
                    servlet_name,
                    "aset_cluster_config_value",
                    key,
                    value,
                    use_servlet_cache=False,
                )
                for servlet_name in await self.aget_all_initialized_servlet_args()
            ]
        )

        return self._cluster_config

    ##############################################
    # Auth cache internal functions
    ##############################################
    async def aresource_access_level(
        self, token: str, resource_uri: str
    ) -> Union[str, None]:
        # If the token in this request matches that of the owner of the cluster,
        # they have access to everything
        config_token = configs.token
        if config_token:
            if config_token == token:
                return ResourceAccess.WRITE

            if resource_uri and rns_client.validate_cluster_token(
                cluster_token=token, cluster_uri=resource_uri
            ):
                return ResourceAccess.WRITE

        return self._auth_cache.lookup_access_level(token, resource_uri)

    async def aget_username(self, token: str) -> str:
        return self._auth_cache.get_username(token)

    async def ahas_resource_access(self, token: str, resource_uri=None) -> bool:
        """Checks whether user has read or write access to a given module saved on the cluster."""
        from runhouse.rns.utils.api import ResourceAccess

        if token is None or self._cluster_name is None:
            # If no token or cluster uri are provided assume no access
            return False

        cluster_access = await self.aresource_access_level(token, self._cluster_name)
        if cluster_access == ResourceAccess.WRITE:
            # if user has write access to cluster will have access to all resources
            return True

        if resource_uri is None and cluster_access not in [
            ResourceAccess.WRITE,
            ResourceAccess.READ,
        ]:
            # If module does not have a name, must have access to the cluster
            return False

        resource_access_level = await self.aresource_access_level(token, resource_uri)
        if resource_access_level not in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return False

        return True

    async def aclear_auth_cache(self, token: str = None):
        self._auth_cache.clear_cache(token)

    ##############################################
    # Key to servlet where it is stored mapping
    ##############################################
    async def aadd_servlet_initialized_args(
        self, servlet_name: str, init_args: CreateProcessParams
    ):
        self._initialized_servlet_args[servlet_name] = init_args

    async def ais_servlet_name_initialized(self, servlet_name: str) -> bool:
        return servlet_name in self._initialized_servlet_args

    async def aget_all_initialized_servlet_args(self) -> Set[str]:
        return self._initialized_servlet_args

    async def aget_key_to_servlet_name_dict_keys(self) -> List[Any]:
        return list(self._key_to_servlet_name.keys())

    async def aget_key_to_servlet_name_dict(self) -> Dict[Any, str]:
        return self._key_to_servlet_name

    async def aget_servlet_name_for_key(self, key: Any) -> str:
        if self.autostop_helper:
            await self.autostop_helper.set_last_active_time_to_now()
        return self._key_to_servlet_name.get(key, None)

    async def aput_servlet_name_for_key(self, key: Any, servlet_name: str):
        if not await self.ais_servlet_name_initialized(servlet_name):
            raise ValueError(
                f"Env servlet name {servlet_name} not initialized, and you tried to mark a resource as in it."
            )
        self._key_to_servlet_name[key] = servlet_name

    async def apop_servlet_name_for_key(self, key: Any, *args) -> str:
        # *args allows us to pass default or not
        return self._key_to_servlet_name.pop(key, *args)

    async def aclear_key_to_servlet_name_dict(self):
        self._key_to_servlet_name = {}

    ##############################################
    # Remove Env Servlet
    ##############################################
    async def aclear_all_references_to_servlet_name(self, servlet_name: str):
        # using lock to prevent status thread access self._initialized_servlet_args before the env is deleted.
        with self.lock:
            del self._initialized_servlet_args[servlet_name]
            deleted_keys = [
                key
                for key, env in self._key_to_servlet_name.items()
                if env == servlet_name
            ]
            for key in deleted_keys:
                self._key_to_servlet_name.pop(key)

            # remove from servlet cache in all processes
            await asyncio.gather(
                *[
                    obj_store.acall_servlet_method(
                        servlet,
                        "aremove_servlet_from_cache",
                        servlet_name,
                    )
                    for servlet in await self.aget_all_initialized_servlet_args()
                ]
            )
            # current process
            await obj_store.adelete_servlet_from_cache(servlet_name)

        return deleted_keys

    ##############################################
    # Periodic Cluster Checks APIs
    ##############################################
    async def asave_status_metrics_to_den(self, status: dict):
        from runhouse.resources.hardware.utils import RunhouseDaemonStatus

        # making a copy so the status won't be modified with pop, since it will be returned after sending to den.
        # (status is passed as pointer).
        status_copy = copy.deepcopy(status)
        servlet_processes = status_copy.pop("env_servlet_processes")

        status_data = {
            "daemon_status": RunhouseDaemonStatus.RUNNING,
            "resource_type": status_copy.get("cluster_config").pop(
                "resource_type", "cluster"
            ),
            "resource_info": status_copy,
            "env_servlet_processes": servlet_processes,
        }

        client = httpx.AsyncClient()

        return await client.post(
            f"{rns_client.api_server_url}/resource/{self._cluster_uri}/cluster/status",
            data=json.dumps(status_data),
            headers=rns_client.request_headers(),
        )

    def save_status_metrics_to_den(self, status: dict):
        return sync_function(self.asave_status_metrics_to_den)(status)

    async def acheck_cluster_status(self, send_to_den: bool = True):
        logger.debug("Performing cluster status checks")
        status, den_resp_status_code = await self.astatus(send_to_den=send_to_den)

        if not send_to_den:
            return status, den_resp_status_code

        if den_resp_status_code == 404:
            logger.info(
                "Cluster has not yet been saved to Den, cannot update status or logs"
            )
        elif den_resp_status_code != 200:
            logger.error("Failed to send cluster status to Den")
        else:
            logger.debug("Successfully sent cluster status to Den")

        return status, den_resp_status_code

    async def acheck_cluster_logs(
        self, interval_size: int = DEFAULT_STATUS_CHECK_INTERVAL
    ):
        logger.debug("Performing logs checks")

        cluster_config = await self.aget_cluster_config()
        prev_end_log_line = cluster_config.get("end_log_line", 0)
        (
            logs_den_resp,
            new_start_log_line,
            new_end_log_line,
        ) = await self.send_cluster_logs_to_den(
            prev_end_log_line=prev_end_log_line,
        )
        if not logs_den_resp:
            logger.debug(
                f"No logs were generated in the past {interval_size} minute(s), logs were not sent to Den"
            )

        elif logs_den_resp.status_code == 200:
            logger.debug("Successfully sent cluster logs to Den")
            await self.aset_cluster_config_value(
                key="start_log_line", value=new_start_log_line
            )
            await self.aset_cluster_config_value(
                key="end_log_line", value=new_end_log_line
            )
        else:
            logger.error("Failed to send logs to Den")

        return logs_den_resp, new_start_log_line, new_end_log_line

    async def aperiodic_autostop_check(self):
        """Periodically check the autostop of the cluster"""
        autostop_interval = int(
            os.getenv("RH_AUTOSTOP_INTERVAL", DEFAULT_AUTOSTOP_CHECK_INTERVAL)
        )
        while True:
            try:
                should_update_autostop: bool = self.autostop_helper is not None
                if should_update_autostop:
                    status, _ = await self.acheck_cluster_status(send_to_den=False)
                    await self._update_autostop(status)
                    logger.debug("Successfully updated autostop")

            except Exception as e:
                logger.error(f"Autostop check has failed: {e}")

                # killing the cluster servlet only if log level is debug
                log_level = os.getenv("RH_LOG_LEVEL") or DEFAULT_LOG_LEVEL
                if log_level.lower() == "debug":
                    kill_actors(gracefully=False, actor_name=self.name)
                    break

            finally:
                logger.debug(f"Autostop interval set to {autostop_interval} seconds")
                await asyncio.sleep(autostop_interval)

    async def aperiodic_cluster_checks(self):
        """Periodically check the status of the cluster, gather metrics about the cluster's utilization & memory,
        and save it to Den."""

        logger.debug("started periodic cluster checks")
        disable_observability = (
            os.getenv("disable_observability", "false").strip().lower() == "true"
        )

        while True:
            should_send_status_and_logs_to_den: bool = (
                configs.token is not None and self._cluster_uri is not None
            )

            if not should_send_status_and_logs_to_den:
                break

            cluster_config = await self.aget_cluster_config()
            interval_size = cluster_config.get(
                "status_check_interval", DEFAULT_STATUS_CHECK_INTERVAL
            )

            try:
                status, den_resp_code = await self.acheck_cluster_status(
                    send_to_den=should_send_status_and_logs_to_den
                )

                if interval_size == -1 or not should_send_status_and_logs_to_den:
                    continue

                logger.debug("Performing cluster checks")

                if disable_observability:
                    logger.info(
                        "Cluster observability not enabled, skipping metrics collection."
                    )
                    break

                if den_resp_code == 404:
                    logger.info(
                        "Cluster has not yet been saved to Den, cannot update status or logs."
                    )
                elif den_resp_code != 200:
                    logger.error(
                        f"Failed to send cluster status to Den, status_code: {den_resp_code}"
                    )
                else:
                    logger.debug("Successfully sent cluster status to Den")

                    (
                        logs_resp_status_code,
                        new_start_log_line,
                        new_end_log_line,
                    ) = await self.acheck_cluster_logs(interval_size=interval_size)

                    if not logs_resp_status_code:
                        logger.debug(
                            f"No logs were generated in the past {interval_size} minute(s), logs were not sent to Den."
                        )

                    elif logs_resp_status_code == 200:
                        await self.aset_cluster_config_value(
                            key="start_log_line", value=new_start_log_line
                        )
                        await self.aset_cluster_config_value(
                            key="end_log_line", value=new_end_log_line
                        )

                        logger.debug("Successfully sent cluster logs to Den.")

            except Exception as e:
                logger.error(f"Cluster checks have failed: {e}.\n")

                # killing the cluster servlet only if log level is debug
                log_level = os.getenv("RH_LOG_LEVEL") or DEFAULT_LOG_LEVEL
                if log_level.lower() == "debug":
                    kill_actors(gracefully=False, actor_name=self.name)

                else:
                    await asyncio.sleep(INCREASED_STATUS_CHECK_INTERVAL)

            finally:
                # make sure that the thread will go to sleep, even if the interval size == -1
                # (meaning that sending status to den is disabled).
                interval_size = (
                    DEFAULT_STATUS_CHECK_INTERVAL
                    if interval_size == -1
                    else interval_size
                )
                await asyncio.sleep(interval_size)

    def periodic_cluster_checks(self):
        # This is only ever called once in its own thread, so we can do asyncio.run here instead of
        # sync_function.
        asyncio.run(self.aperiodic_cluster_checks())

    def periodic_autostop_check(self):
        asyncio.run(self.aperiodic_autostop_check())

    async def _update_autostop(self, status: dict):
        function_running = any(
            any(
                len(
                    resource["env_resource_mapping"][resource_name].get(
                        "active_function_calls", []
                    )
                )
                > 0
                for resource_name in resource["env_resource_mapping"].keys()
            )
            for resource in status.get("env_servlet_processes", {}).values()
        )
        if function_running:
            await self.autostop_helper.set_last_active_time_to_now()

        # We do this separately from the set_last_active_time_to_now call above because
        # function_running will only reflect activity from functions which happen to be running during
        # the status check. We still need to attempt to register activity for functions which have
        # been called and completed.
        await self.autostop_helper.register_activity_if_needed()

    async def _status_for_servlet(self, servlet_name):
        try:
            (
                objects_in_servlet,
                servlet_utilization_data,
            ) = await obj_store.acall_servlet_method(
                servlet_name, method="astatus_local"
            )

            return {
                "servlet_name": servlet_name,
                "objects_in_servlet": objects_in_servlet,
                "servlet_utilization_data": servlet_utilization_data,
            }

        # Need to catch the exception here because we're running this in a gather,
        # and need to know which env servlet failed
        except Exception as e:
            return {"servlet_name": servlet_name, "Exception": e}

    async def _aperiodic_gpu_check(self):
        """periodically collects cluster gpu usage"""
        import pynvml

        logger.debug("Started gpu usage collection")

        pynvml.nvmlInit()  # init nvidia ml info collection

        while True:
            try:
                gpu_count = pynvml.nvmlDeviceGetCount()
                with self.lock:
                    if not self.gpu_metrics:
                        self.gpu_metrics = {device: [] for device in range(gpu_count)}

                    for gpu_index in range(gpu_count):
                        handle = pynvml.nvmlDeviceGetHandleByIndex(gpu_index)
                        util_info = pynvml.nvmlDeviceGetUtilizationRates(handle)
                        memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)

                        total_memory = memory_info.total  # in bytes
                        used_memory = memory_info.used  # in bytes
                        free_memory = memory_info.free  # in bytes
                        utilization_percent = util_info.gpu / 1.0  # make it float

                        # to reduce cluster memory usage (we are saving the gpu_usage info on the cluster),
                        # we save only the most updated gpu usage. If for some reason the size of updated_gpu_info is
                        # too big, we remove the older gpu usage info.
                        # This is relevant when using cluster.status() directly and not relying on status being sent to den.
                        updated_gpu_info = self.gpu_metrics[gpu_index]
                        if len(updated_gpu_info) + 1 > MAX_GPU_INFO_LEN:
                            updated_gpu_info = updated_gpu_info[REDUCED_GPU_INFO_LEN:]
                        updated_gpu_info.append(
                            {
                                "total_memory": total_memory,
                                "used_memory": used_memory,
                                "free_memory": free_memory,
                                "utilization_percent": utilization_percent,
                            }
                        )
                        self.gpu_metrics[gpu_index] = updated_gpu_info

            except Exception as e:
                logger.error(
                    f"{self._cluster_name}'s GPU metrics collection failed: {str(e)}"
                )

                pynvml.nvmlShutdown()

                # killing the cluster servlet only if log level is debug
                log_level = os.getenv("RH_LOG_LEVEL") or DEFAULT_LOG_LEVEL
                if log_level.lower() == "debug":
                    kill_actors(gracefully=False, actor_name=self.name)
                else:
                    # increase sleep interval
                    await asyncio.sleep(INCREASED_GPU_COLLECTION_INTERVAL)

            finally:
                # collects gpu usage every 5 seconds.
                await asyncio.sleep(GPU_COLLECTION_INTERVAL)

    def _periodic_gpu_check(self):
        # This is only ever called once in its own thread, so we can do asyncio.run here instead of `sync_function`.
        asyncio.run(self._aperiodic_gpu_check())

    def _get_node_gpu_usage(self, server_pid: int):
        # TODO [SB] currently works correctly for a single node GPU. Multinode-clusters will be supported shortly.
        collected_gpus_info = copy.deepcopy(self.gpu_metrics)

        if collected_gpus_info is None or not collected_gpus_info[0]:
            return None

        cluster_gpu_usage = get_gpu_usage(
            collected_gpus_info=collected_gpus_info, servlet_type=ServletType.cluster
        )

        # will be useful for multi-node clusters.
        cluster_gpu_usage["server_pid"] = server_pid

        return cluster_gpu_usage

    async def astatus(self, send_to_den: bool = False) -> Tuple[Dict, Optional[int]]:
        import psutil

        config_cluster = copy.deepcopy(self._cluster_config)

        # Popping out creds because we don't want to show them in the status
        config_cluster.pop("creds", None)

        # Getting data from each env servlet about the objects it contains and the utilization data
        servlet_utilization_data = {}
        with self.lock:
            servlets_status = await asyncio.gather(
                *[
                    self._status_for_servlet(servlet_name)
                    for servlet_name in await self.aget_all_initialized_servlet_args()
                ],
            )

        # Store the data for the appropriate env servlet name
        for env_status in servlets_status:
            servlet_name = env_status.get("servlet_name")

            # Nothing to store if there was an exception
            if "Exception" in env_status.keys():
                e = env_status.get("Exception")
                logger.warning(
                    f"Exception {str(e)} in status for env servlet {servlet_name}"
                )
                servlet_utilization_data[servlet_name] = {}

            else:
                # Store what was in the env and the utilization data
                env_memory_info = env_status.get("servlet_utilization_data")
                env_memory_info["env_resource_mapping"] = env_status.get(
                    "objects_in_servlet"
                )
                servlet_utilization_data[servlet_name] = env_memory_info

        # TODO: decide if we need this info at all: cpu_usage, memory_usage, disk_usage
        cpu_utilization = psutil.cpu_percent(interval=0)

        # A dictionary that match the keys of psutil.virtual_memory()._asdict() to match the keys we expect in Den.
        relevant_memory_info = {
            "available": "free_memory",
            "percent": "percent",
            "total": "total_memory",
            "used": "used_memory",
        }

        # Fields: `total`, `available`, `percent`, `used`, `free`, `active`, `inactive`, `buffers`, `cached`, `shared`, `slab`
        # according to psutil docs, percent = (total - available) / total * 100
        memory_usage = psutil.virtual_memory()._asdict()

        memory_usage = {
            relevant_memory_info[k]: memory_usage[k]
            for k in relevant_memory_info.keys()
        }

        # get general gpu usage
        server_gpu_usage = (
            self._get_node_gpu_usage(self.pid)
            if self._cluster_config.get("is_gpu")
            else None
        )
        gpu_utilization = (
            server_gpu_usage.pop("utilization_percent", None)
            if server_gpu_usage
            else None
        )

        # rest the gpu_info only after the status was sent to den. If we should not send status to den,
        # self.gpu_metrics will not be updated at all, therefore should not be reset.
        if send_to_den:
            with self.lock:
                self.gpu_metrics = None

        status_data = {
            "cluster_config": config_cluster,
            "runhouse_version": runhouse.__version__,
            "server_pid": self.pid,
            "env_servlet_processes": servlet_utilization_data,
            "server_cpu_utilization": cpu_utilization,
            "server_gpu_utilization": gpu_utilization,
            "server_memory_usage": memory_usage,
            "server_gpu_usage": server_gpu_usage,
        }

        # converting status_data to ResourceStatusData instance to verify we constructed the status data correctly
        status_data = ResourceStatusData(**status_data).model_dump()

        if send_to_den:
            logger.debug("Sending cluster status to Den")
            den_resp = self.save_status_metrics_to_den(status=status_data)
            return status_data, den_resp.status_code

        return status_data, None

    def status(self, send_to_den: bool = False):
        return sync_function(self.astatus)(send_to_den=send_to_den)

    ##############################################
    # Save cluster logs to Den
    ##############################################
    def _get_logs(self):
        from pathlib import Path

        if not Path(SERVER_LOGFILE).exists():
            if not self._is_log_file_ready:
                self._is_log_file_ready = True
                return ""
            raise ValueError(f"{SERVER_LOGFILE} is not found, can't send logs to den")

        with open(SERVER_LOGFILE) as log_file:
            log_lines = log_file.readlines()
        cleaned_log_lines = [ColoredFormatter.format_log(line) for line in log_lines]
        return " ".join(cleaned_log_lines)

    def _generate_logs_file_name(self):
        current_timestamp = datetime.datetime.utcnow().strftime("%Y_%m_%d_%H_%M_%S")
        return f"{current_timestamp}_{SERVER_LOGS_FILE_NAME}"

    async def send_cluster_logs_to_den(
        self, prev_end_log_line: int
    ) -> Tuple[Optional[requests.Response], Optional[int], Optional[int]]:
        """Load the most recent logs from the server's log file and send them to Den."""
        # setting to a list, so it will be easier to get the end line num  + the logs delta to send to den.
        latest_logs = self._get_logs().split("\n")

        # minus 1 because we start counting logs from 0.
        new_end_log_line = len(latest_logs) - 1

        if new_end_log_line < prev_end_log_line:
            # Likely a sign that the daemon was restarted, so we should start from the beginning
            prev_end_log_line = 0

        # len(latest_logs) may equal to 0 if we are in the process of setting up the cluster servlet.
        # In this case the server.log might not have been created yet, therefore self._get_logs() returns an empty string ("").
        logs_to_den = (
            "\n".join(latest_logs[prev_end_log_line:])
            if len(latest_logs) > 0
            else latest_logs
        )

        if len(logs_to_den) == 0:
            return None, None, None

        logs_data = {
            "file_name": self._generate_logs_file_name(),
            "logs": logs_to_den,
            "start_line": prev_end_log_line,
            "end_line": new_end_log_line,
        }

        post_logs_resp = requests.post(
            f"{rns_client.api_server_url}/resource/{self._cluster_uri}/logs",
            data=json.dumps(logs_data),
            headers=rns_client.request_headers(),
        )

        resp_status_code = post_logs_resp.status_code
        if resp_status_code != 200:
            logger.error(
                f"{resp_status_code}: Failed to send cluster logs to Den: {post_logs_resp.json()}"
            )

        return post_logs_resp, prev_end_log_line, new_end_log_line



from concurrent.futures.thread import ThreadPoolExecutor
from typing import List

from runhouse.resources.distributed.supervisor import Supervisor

from runhouse.resources.hardware import Cluster, OnDemandCluster

from runhouse.resources.module import Module


class PyTorchDistributed(Supervisor):
    def __init__(self, name, replicas: List[Module] = None, port=None, **kwargs):
        super().__init__(name=name, **kwargs)
        self._replicas = replicas or []
        self._port = port

    def _compute_signature(self, rich=False):
        return self.local._replicas[0].signature(rich=rich)

    def _find_available_port_on_head_node(self):
        find_available_port_cmd = "python -c \"import socket; s=socket.socket(); s.bind(('', 0)); print(s.getsockname()[1]); s.close()\""
        status_code, stdout, _ = self._replicas[0].system.run_bash(
            find_available_port_cmd,
            node=self._replicas[0].system.head_ip,
            require_outputs=True,
        )

        if status_code != 0:
            raise RuntimeError(f"Failed to find available port on head rank: {stdout}")
        return stdout

    def forward(self, item, *args, **kwargs):
        port = self._port or self._find_available_port_on_head_node()

        def run_on_replica(replica, rank):
            # Per https://pytorch.org/docs/stable/distributed.html#environment-variable-initialization
            master_addr = (
                self.system.internal_ips[0]
                if isinstance(self.system, OnDemandCluster)
                else self.system.head_ip
                if isinstance(self.system, Cluster)
                else "localhost"
            )
            dist_config = {
                "MASTER_ADDR": master_addr,
                "MASTER_PORT": port,
                "RANK": str(rank),
                "WORLD_SIZE": str(len(self._replicas)),
            }
            replica.system.set_process_env_vars(replica.process, dist_config)
            method = getattr(replica, item)
            return method(*args, **kwargs)

        with ThreadPoolExecutor(max_workers=len(self._replicas)) as executor:
            res = executor.map(
                run_on_replica, self._replicas, range(len(self._replicas))
            )
            res = list(res)

        return res

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)



import json
import subprocess

import pytest
import requests

import runhouse as rh
from runhouse.globals import rns_client
from runhouse.servers.http.certs import TLSCertConfig

from tests.utils import friend_account


class TestResourceSharing:
    UNIT = {"resource": ["shared_function"]}
    LOCAL = {"resource": ["shared_function"]}
    MINIMAL = {"resource": ["shared_function"]}
    RELEASE = {"resource": ["shared_function"]}
    MAXIMAL = {"resource": ["shared_function"]}

    @staticmethod
    def call_func_with_curl(cluster, func_name, cluster_token, **kwargs):
        from urllib.parse import quote

        path_to_cert = TLSCertConfig().cert_path

        # Example for using the cert for verification:
        # curl --cacert '/Users/josh.l/.rh/certs/rh_server.crt' https://localhost:8444/check
        # >> {"rh_version":"0.0.18"}

        query_string = "&".join(
            [f"{quote(str(k))}={quote(str(v))}" for k, v in kwargs.items()]
        )

        cmd = f"""curl --cacert {path_to_cert} -X GET "{cluster.endpoint()}/{func_name}/call?{query_string}" -H "Content-Type: application/json" -H "Authorization: Bearer {cluster_token}" """  # noqa

        res = subprocess.run(
            cmd,
            shell=True,
            check=True,
            capture_output=True,
            text=True,
        )
        return res

    @staticmethod
    def call_cluster_methods(cluster, valid_token):
        cluster_methods = [
            (cluster.put, ("test_obj", list(range(10)))),
            (cluster.get, ("test_obj",)),
            (cluster.keys, ()),
            (cluster.rename, ("test_obj", "test_obj2")),
        ]

        for method, args in cluster_methods:
            try:
                method(*args)
            except Exception as e:
                if valid_token:
                    raise e
                else:
                    assert "Error calling" in str(e)

    @pytest.mark.level("local")
    def test_calling_shared_resource(self, resource):
        cluster = resource.system

        # Run commands on cluster with current token
        return_codes = cluster.run_python(["import numpy", "print(numpy.__version__)"])
        assert return_codes[0][0] == 0

        # Call function with current token via CURL
        cluster_token = rns_client.cluster_token(cluster.rns_address)
        res = self.call_func_with_curl(
            cluster, resource.name, cluster_token, **{"a": 1, "b": 2}
        )

        # Note: kwarg vals should no longer be concatenated once schema support is addde
        assert "12" in res.stdout

        # Reload the shared function and call it
        reloaded_func = rh.function(name=resource.rns_address)
        assert reloaded_func(1, 2) == 3

    @pytest.mark.level("local")
    def test_use_cluster_apis_for_shared_resource(self, resource):
        # Should be able to use the shared cluster APIs if given access
        current_token = rh.configs.token

        # Confirm we can perform cluster actions with the current token
        cluster = resource.system
        self.call_cluster_methods(cluster, valid_token=True)

        # Should not be able to install packages on the cluster with read access
        try:
            cluster.install_packages(["numpy", "pandas"])
        except Exception as e:
            assert "No read or write access to requested resource" in str(e)

        # Confirm we cannot perform actions on the cluster with an invalid token
        rh.configs.token = "abc123"
        self.call_cluster_methods(cluster, valid_token=False)

        # Reset back to valid token
        rh.configs.token = current_token

    @pytest.mark.level("local")
    def test_use_resource_apis(self, resource):
        current_token = rh.configs.token
        cluster = resource.system

        # Call the function with current valid token
        assert resource(2, 2) == 4

        # Reload the shared function and call it
        reloaded_func = rh.function(name=resource.rns_address)
        assert reloaded_func(1, 2) == 3

        # Use invalid token to confirm no function access
        rh.configs.token = "abc123"
        with pytest.raises(Exception):
            # cluster will throw error since the cluster token is invalid
            # note: use the "call" method directly in order to pass new request headers with invalid token
            cluster._http_client.call(
                key=reloaded_func.name,
                method_name="call",
                headers=rns_client.request_headers(resource.rns_address),
            )

        # Reset back to valid token and confirm we can call function again
        rh.configs.token = current_token
        cluster_token = rns_client.cluster_token(cluster.rns_address)

        res = self.call_func_with_curl(
            cluster, resource.name, cluster_token, **{"a": 1, "b": 2}
        )
        assert "12" in res.stdout

    @pytest.mark.level("local")
    def test_calling_resource_with_cluster_read_access(self, resource):
        """Check that a user with read access to a cluster can call a function on that cluster, even without having
        explicit access to the function."""
        current_username = rh.configs.username
        cluster = resource.system

        cluster_uri = rns_client.resource_uri(cluster.rns_address)
        resource_uri = rns_client.resource_uri(resource.rns_address)

        with friend_account():
            friend_account_request_headers = rns_client.request_headers()
            # Give user write access to cluster from test account
            resp = requests.put(
                f"{rns_client.api_server_url}/resource/{cluster_uri}/users/access",
                data=json.dumps(
                    {
                        "users": [current_username],
                        "access_level": "read",
                        "notify_users": False,
                    }
                ),
                headers=friend_account_request_headers,
            )
            if resp.status_code != 200:
                assert (
                    False
                ), f"Failed to grant user {current_username} write access to the cluster: {resp.text}"

            # Delete user access to function
            requests.delete(
                f"{rns_client.api_server_url}/resource/{resource_uri}/user/{current_username}",
                headers=friend_account_request_headers,
            )

        # Confirm user can still call the function with write access to the cluster
        cluster_token = rns_client.cluster_token(cluster.rns_address)
        res = self.call_func_with_curl(
            cluster,
            resource.name,
            cluster_token,
            **{"a": 1, "b": 2, "serialization": "none"},
        )

        assert "12" in res.stdout

        assert resource(1, 2) == 3

    @pytest.mark.level("local")
    @pytest.mark.skip("Breaking in CI - investigate further")
    def test_calling_resource_with_no_cluster_access(self, resource):
        """Check that a user with no access to the cluster can still call a function on that cluster if they were
        given explicit access to the function."""
        current_username = rh.configs.username
        cluster = resource.system

        cluster_uri = rns_client.resource_uri(cluster.rns_address)

        with friend_account():
            # Delete user access to cluster using the test account
            resp = requests.delete(
                f"{rns_client.api_server_url}/resource/{cluster_uri}/user/{current_username}",
                headers=rns_client.request_headers(),
            )
            if resp.status_code != 200:
                assert (
                    False
                ), f"Failed to remove access to the cluster for user: {current_username}: {resp.text}"

        # Confirm current user can still call the function (which they still have explicit access to)
        cluster_token = rns_client.cluster_token(cluster.rns_address)
        res = self.call_func_with_curl(
            cluster, resource.name, cluster_token, **{"a": 1, "b": 2}
        )
        assert "12" in res.stdout

        assert resource(1, 2) == 3



import logging
from pathlib import Path
from typing import Dict, Optional

from runhouse.utils import run_setup_command

from .package import Package


class GitPackage(Package):
    RESOURCE_TYPE = "package"

    def __init__(
        self,
        name: str = None,
        git_url: str = None,
        install_method: str = None,
        install_args: str = None,
        revision: str = None,
        dryrun: bool = False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Github Package resource.

        .. note::
            To create a git package, please use the factory method :func:`git_package` or :func:`package`.
        """
        super().__init__(
            name=name,
            dryrun=dryrun,
            install_method=install_method,
            install_target="./" + git_url.split("/")[-1].replace(".git", ""),
            install_args=install_args,
        )
        self.git_url = git_url
        self.revision = revision

    def config(self, condensed: bool = True):
        # If the package is just a simple Package.from_string string, no
        # need to store it in rns, just give back the string.
        # if self.install_method in ['pip', 'conda', 'git']:
        #     return f'{self.install_method}:{self.name}'
        config = super().config(condensed)
        self.save_attrs_to_config(config, ["git_url", "revision"])
        return config

    def __str__(self):
        if self.name:
            return f"GitPackage: {self.name}"
        return f"GitPackage: {self.git_url}@{self.revision}"

    # TODO for cluster
    def _install(self, cluster: "Cluster" = None, conda_env_name: Optional[str] = None):
        from runhouse.resources.folders import Folder, folder

        if cluster and isinstance(self.install_target, str):
            install_target = folder(path=self.install_target, system=cluster)
        else:
            install_target = self.install_target

        # Clone down the repo
        if (cluster and not install_target.exists_in_system()) or (
            not cluster and not Path(self.install_target).exists()
        ):
            logging.info(f"Cloning: git clone {self.git_url}")
            run_setup_command(f"git clone {self.git_url}", cluster=cluster)
        else:
            install_path = (
                install_target.path
                if isinstance(install_target, Folder)
                else install_target
            )
            run_setup_command(
                f"git -C {install_path} fetch {self.git_url}", cluster=cluster
            )

        if self.revision:
            logging.info(f"Checking out revision: git checkout {self.revision}")
            run_setup_command(f"git -C {install_target} checkout {self.revision}")

        # Use super to install the package
        super()._install(cluster=cluster, conda_env_name=conda_env_name)

    @staticmethod
    def from_config(config: Dict, dryrun: bool = False, _resolve_children: bool = True):
        return GitPackage(**config, dryrun=dryrun)


def git_package(
    name: str = None,
    git_url: str = None,
    revision: str = None,
    install_method: str = None,
    install_str: str = None,
    load_from_den: bool = True,
    dryrun: bool = False,
):
    """
    Builds an instance of :class:`GitPackage`.

    Args:
        name (str, optional): Name to assign the package resource.
        git_url (str, optional): The GitHub URL of the package to install.
        revision (str, optional): Version of the Git package to install.
        install_method (str, optional): Method for installing the package. If left blank, defaults to
            local installation.
        install_str (str, optional): Additional arguments to add to installation command.
        load_from_den (bool, optional): Whether to try loading the package from Den. (Default: ``True``)
        dryrun (bool, optional): Whether to load the Package object as a dryrun, or create the Package if
            it doesn't exist. (Default: ``False``)

    Returns:
        GitPackage: The resulting GitHub Package.

    Example:
        >>> rh.git_package(git_url='https://github.com/runhouse/runhouse.git',
        >>>               install_method='pip', revision='v0.0.1')

    """
    if name and not any([install_method, install_str, git_url, revision]):
        # If only the name is provided and dryrun is set to True
        return Package.from_name(name, load_from_den=load_from_den, dryrun=dryrun)

    install_method = install_method or "local"
    if git_url is not None:
        if not git_url.endswith(".git"):
            git_url += ".git"

    return GitPackage(
        git_url=git_url,
        revision=revision,
        install_method=install_method,
        install_args=install_str,
        dryrun=dryrun,
    )



import ast
import datetime
import functools
import json
import os
import uuid
from enum import Enum
from typing import Tuple

from requests import Response


def timing(func):
    def wrapper(*args, **kwargs):
        import time

        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f"Finished {func.__name__.title()} in {int((end - start))} seconds")
        return result

    return wrapper


def remove_null_values_from_dict(source_dic: dict) -> dict:
    return {k: v for k, v in source_dic.items() if v is not None}


def load_resp_content(resp: Response) -> dict:
    return json.loads(resp.content)


def read_resp_data(resp: Response):
    return load_resp_content(resp).get("data", {})


def to_bool(value):
    try:
        return ast.literal_eval(value)
    except:
        return value


def is_jsonable(myjson):
    try:
        json.dumps(myjson)
    except (TypeError, OverflowError):
        return False
    return True


def generate_uuid():
    return uuid.uuid4().hex


def utc_now():
    """Current time as datetime object."""
    return datetime.datetime.now(datetime.timezone.utc)


def log_timestamp():
    """Return as timestamp in nanoseconds."""
    return int(utc_now().timestamp() * 1e9)


def log_datetime():
    """Current time as readable datetime string.
    Example: '2023-04-23'"""
    return utc_now().strftime("%Y-%m-%d")


def resolve_absolute_path(path: str):
    return os.path.abspath(os.path.expanduser(path))


def relative_file_path(file_path: str):
    """Convert to a relative path if it is not already one."""
    if file_path.startswith("~"):
        return file_path

    # Convert to a relative path
    relative_path = os.path.relpath(file_path, os.path.expanduser("~"))
    relative_path = relative_path.replace("\\", "/")

    if not relative_path.startswith("~"):
        relative_path = f"~/{relative_path}"

    return relative_path


def generate_ssh_keys() -> Tuple[str, str]:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    # Adapted from: https://github.com/skypilot-org/skypilot/blob/2c7419cdd22adb440bd37702be6f1ddc1c287684/sky/authentication.py#L106 # noqa
    private_key_path = os.path.expanduser("~/.ssh/sky-key")
    public_key_path = os.path.expanduser("~/.ssh/sky-key.pub")

    if os.path.exists(private_key_path):
        return private_key_path, public_key_path

    key = rsa.generate_private_key(
        backend=default_backend(), public_exponent=65537, key_size=2048
    )

    private_key = (
        key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
        .decode("utf-8")
        .strip()
    )

    public_key = (
        key.public_key()
        .public_bytes(
            serialization.Encoding.OpenSSH, serialization.PublicFormat.OpenSSH
        )
        .decode("utf-8")
        .strip()
    )

    key_dir = os.path.dirname(private_key_path)
    os.makedirs(key_dir, exist_ok=True, mode=0o700)

    with open(
        private_key_path,
        "w",
        encoding="utf-8",
        opener=functools.partial(os.open, mode=0o600),
    ) as f:
        f.write(private_key)

    with open(
        public_key_path,
        "w",
        encoding="utf-8",
        opener=functools.partial(os.open, mode=0o644),
    ) as f:
        f.write(public_key)

    return private_key_path, public_key_path


class ResourceAccess(str, Enum):
    WRITE = "write"
    READ = "read"
    DENIED = "denied"


class ResourceVisibility(str, Enum):
    PRIVATE = "private"
    UNLISTED = "unlisted"
    PUBLIC = "public"



import os

import pytest

import runhouse as rh

from runhouse.constants import DEFAULT_PROCESS_NAME
from runhouse.globals import rns_client

import tests.test_resources.test_resource

from tests.utils import friend_account

_provider_path_map = {
    "aws": "credentials",
    "gcp": "credentials.json",
    "azure": "clouds.config",
    "lambda": "lambda_key",
    "github": "hosts.yml",
    "huggingface": "token",
    "kubernetes": "config",
    "ssh": "id_rsa",
    "sky": "sky-key",
    "custom_provider": "~/.rh/tests/custom_provider/config.json",
}

provider_secrets = [
    "aws_secret",
    "azure_secret",
    "gcp_secret",
    "lambda_secret",
    "github_secret",
    "huggingface_secret",
    "kubeconfig_secret",
    "ssh_secret",
    "sky_secret",
    "custom_provider_secret",
]

api_secrets = [
    "anthropic_secret",
    "cohere_secret",
    "langchain_secret",
    "openai_secret",
    "pinecone_secret",
    "wandb_secret",
]


def assert_delete_local(secret, contents: bool = False):
    secret.delete(contents=contents) if contents else secret.delete()
    with pytest.raises(Exception) as e:
        rh.secret(name=secret.name)
        assert isinstance(e, Exception)
    if contents and secret.path:
        assert not os.path.exists(os.path.expanduser(secret.path))


def _get_env_var_value(env_var):
    import os

    return os.environ[env_var]


@pytest.mark.secrettest
class TestSecret(tests.test_resources.test_resource.TestResource):
    MAP_FIXTURES = {"resource": "secret"}

    UNIT = {"secret": ["test_secret"] + provider_secrets + api_secrets}
    LOCAL = {
        "secret": ["test_secret"] + provider_secrets + api_secrets,
        "cluster": [
            "docker_cluster_pk_ssh_no_auth",
        ],
    }
    MINIMAL = {
        "secret": [
            "test_secret",
            "aws_secret",
            "ssh_secret",
            "openai_secret",
            "custom_provider_secret",
        ],
        "cluster": ["local_launched_ondemand_aws_docker_cluster"],
    }
    RELEASE = {
        "secret": ["test_secret"] + provider_secrets,
        "cluster": [
            "local_launched_ondemand_aws_docker_cluster",
            "static_cpu_pwd_cluster",
        ],
    }
    MAXIMAL = {
        "secret": ["test_secret"] + provider_secrets,
        "cluster": [
            "local_launched_ondemand_aws_docker_cluster",
            "ondemand_gcp_cluster",
            "ondemand_k8s_cluster",
            "ondemand_k8s_docker_cluster",
            "ondemand_aws_https_cluster_with_auth",
            "static_cpu_pwd_cluster",
            "multinode_cpu_docker_conda_cluster",
            "docker_cluster_pk_ssh_no_auth",
            "docker_cluster_pwd_ssh_no_auth",
            "docker_cluster_pk_ssh_den_auth",
        ],
    }

    @pytest.mark.level("unit")
    def test_secret_factory_and_properties(self, secret):
        assert isinstance(secret, rh.Secret)

    @pytest.mark.level("local")
    def test_provider_secret_to_cluster_values(self, secret, cluster):
        remote_secret = secret.to(cluster)
        assert cluster.get(remote_secret.name)
        assert cluster.get(remote_secret.name).values
        cluster.delete(remote_secret.name)

    @pytest.mark.level("local")
    def test_provider_secret_to_cluster_path(self, secret, cluster):
        if not isinstance(secret, rh.ProviderSecret):
            return

        if secret.name not in _provider_path_map.keys():
            return

        test_path = os.path.join("~/.rh/tests", _provider_path_map[secret.provider])
        remote_secret = secret.to(cluster, path=test_path)

        assert remote_secret.path.system == cluster
        assert remote_secret.values == secret.values
        assert remote_secret.path.exists_in_system()

        delete_contents = secret.provider not in ["ssh", "sky"]
        remote_secret.delete(contents=delete_contents)
        assert_delete_local(secret, contents=delete_contents)

    @pytest.mark.level("local")
    def test_provider_secret_to_cluster_process(self, secret, cluster):
        if not isinstance(secret, rh.ProviderSecret):
            return

        env_vars = secret._DEFAULT_ENV_VARS
        if not env_vars:
            return

        get_remote_val = rh.function(_get_env_var_value, name="get_env_vars").to(
            cluster
        )
        secret.to(cluster, process=DEFAULT_PROCESS_NAME)

        for (key, val) in env_vars.items():
            assert get_remote_val(val) == secret.values[key]

    @pytest.mark.level("unit")
    def test_share_and_revoke_secret(self, test_secret):
        vault_secret = test_secret.save(f"{test_secret.name}_shared")
        vault_secret.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        with friend_account():
            reloaded_secret = rh.secret(name=vault_secret.rns_address)
            assert reloaded_secret.values == test_secret.values

        vault_secret.revoke(users=["info@run.house"])
        with pytest.raises(Exception):
            with friend_account():
                rh.secret(name=vault_secret.rns_address)

    @pytest.mark.level("unit")
    def test_sharing_public_secret(self, test_secret):
        # Create & share
        with friend_account():
            test_headers = rns_client.request_headers()
            vault_secret = rh.secret(name=test_secret.name, values=test_secret.values)
            vault_secret.save(headers=test_headers)

            rns_address = vault_secret.rns_address

            # Make the resource available to all users, without explicitly sharing with any users
            vault_secret.share(visibility="public", notify_users=False)

            del vault_secret

        # By default we can re-load the public resource
        reloaded_secret = rh.secret(name=rns_address)

        # NOTE: currently not loading the values for public secret resources (i.e. reloaded_secret.values will be empty)
        assert reloaded_secret

    @pytest.mark.level("local")
    def test_sync_secrets(self, secret, cluster):
        if not isinstance(secret, rh.ProviderSecret):
            return

        if secret.provider in _provider_path_map.keys():
            test_path = _provider_path_map[secret.provider]
            secret = secret.write(path=test_path)
            cluster.sync_secrets([secret])

            remote_folder = rh.folder(path=secret.path, system=cluster)
            assert remote_folder.exists_in_system()
            assert secret._from_path(remote_folder.path) == secret.values

            assert_delete_local(secret, contents=True)
            remote_folder.rm()
        else:
            cluster.sync_secrets([secret])
            assert cluster.get(secret.name)



import logging
import os
import shlex
from typing import Dict, List, Optional, Tuple, Union

from runhouse.constants import DEFAULT_DOCKER_CONTAINER_NAME

from runhouse.logger import get_logger

from runhouse.resources.hardware.sky import log_lib, subprocess_utils

from runhouse.resources.hardware.sky.command_runner import (
    _DEFAULT_CONNECT_TIMEOUT,
    KubernetesCommandRunner,
    ssh_options_list,
    SSHCommandRunner,
    SshMode,
)

from runhouse.resources.hardware.utils import _ssh_base_command

logger = get_logger(__name__)

# Get rid of the constant "Found credentials in shared credentials file: ~/.aws/credentials" message
try:
    import boto3

    boto3.set_stream_logger(name="botocore.credentials", level=logging.ERROR)
except ImportError:
    pass


def get_docker_user(cluster: "Cluster", ssh_creds: Dict) -> str:
    """Find docker container username."""
    runner = SkySSHRunner(
        node=(cluster.head_ip, cluster.ssh_port),
        ssh_user=ssh_creds.get("ssh_user", None),
        ssh_private_key=ssh_creds.get("ssh_private_key", None),
        ssh_control_name=ssh_creds.get(
            "ssh_control_name", f"{cluster.head_ip}:{cluster.ssh_port}"
        ),
    )
    container_name = DEFAULT_DOCKER_CONTAINER_NAME
    whoami_returncode, whoami_stdout, whoami_stderr = runner.run(
        f"sudo docker exec {container_name} whoami",
        stream_logs=True,
        require_outputs=True,
    )
    assert whoami_returncode == 0, (
        f"Failed to get docker container user. Return "
        f"code: {whoami_returncode}, Error: {whoami_stderr}"
    )
    docker_user = whoami_stdout.strip()
    logger.debug(f"Docker container user: {docker_user}")
    return docker_user


class SkySSHRunner(SSHCommandRunner):
    def __init__(
        self,
        node: Tuple[str, int],
        ssh_user: Optional[str] = None,
        ssh_private_key: Optional[str] = None,
        ssh_control_name: Optional[str] = "__default__",
        ssh_proxy_command: Optional[str] = None,
        docker_user: Optional[str] = None,
        disable_control_master: Optional[bool] = False,
        local_bind_port: Optional[int] = None,
        use_docker_exec: Optional[bool] = False,
    ):
        super().__init__(
            node,
            ssh_user,
            ssh_private_key,
            ssh_control_name,
            ssh_proxy_command,
            docker_user,
            disable_control_master,
        )

        # RH modified
        self.docker_user = docker_user
        self.local_bind_port = local_bind_port
        self.remote_bind_port = None
        self.use_docker_exec = use_docker_exec

    def _ssh_base_command(
        self,
        *,
        ssh_mode: SshMode,
        port_forward: Optional[List[int]],
        connect_timeout: Optional[int] = None,
    ) -> List[str]:
        docker_ssh_proxy_command = (
            self._docker_ssh_proxy_command(["ssh"])
            if self._docker_ssh_proxy_command
            else None
        )

        return _ssh_base_command(
            address=self.ip,
            ssh_user=self.ssh_user,
            ssh_private_key=self.ssh_private_key,
            ssh_control_name=self.ssh_control_name,
            ssh_proxy_command=self._ssh_proxy_command,
            ssh_port=self.port,
            docker_ssh_proxy_command=docker_ssh_proxy_command,
            disable_control_master=self.disable_control_master,
            ssh_mode=ssh_mode,
            port_forward=port_forward,
            connect_timeout=connect_timeout,
        )

    def run(
        self,
        cmd: Union[str, List[str]],
        *,
        require_outputs: bool = False,
        port_forward: Optional[List[int]] = None,
        # Advanced options.
        log_path: str = os.devnull,
        # If False, do not redirect stdout/stderr to optimize performance.
        process_stream: bool = True,
        stream_logs: bool = True,
        ssh_mode: SshMode = SshMode.NON_INTERACTIVE,
        separate_stderr: bool = False,
        connect_timeout: Optional[int] = None,
        source_bashrc: bool = True,  # RH MODIFIED
        skip_lines: int = 0,
        return_cmd: bool = False,  # RH MODIFIED
        quiet_ssh: bool = False,  # RH MODIFIED
        **kwargs,
    ) -> Union[int, Tuple[int, str, str]]:
        """Uses 'ssh' to run 'cmd' on a node with ip.

        Args:
            ip: The IP address of the node.
            cmd: The command to run.
            port_forward: A list of ports to forward from the localhost to the
            remote host.

            Advanced options:

            require_outputs: Whether to return the stdout/stderr of the command.
            log_path: Redirect stdout/stderr to the log_path.
            stream_logs: Stream logs to the stdout/stderr.
            check: Check the success of the command.
            ssh_mode: The mode to use for ssh.
                See SSHMode for more details.
            separate_stderr: Whether to separate stderr from stdout.
                        connect_timeout: timeout in seconds for the ssh connection.
            source_bashrc: Whether to source the bashrc before running the
                command.
            skip_lines: The number of lines to skip at the beginning of the
                output. This is used when the output is not processed by
                SkyPilot but we still want to get rid of some warning messages,
                such as SSH warnings.
            return_cmd: If True, return the command string instead of running it.
            quiet_ssh: If True, do not print the OpenSSH outputs (i.e. add "-q" option to ssh).


        Returns:
            returncode
            or
            A tuple of (returncode, stdout, stderr).
        """
        base_ssh_command = self._ssh_base_command(
            ssh_mode=ssh_mode,
            port_forward=port_forward,
            connect_timeout=connect_timeout,
        )
        if ssh_mode == SshMode.LOGIN:
            assert isinstance(cmd, list), "cmd must be a list for login mode."
            command = base_ssh_command + cmd
            proc = subprocess_utils.run(command, shell=False, check=False)
            return proc.returncode, "", ""

        if quiet_ssh:  # RH MODIFIED
            base_ssh_command.append("-q")

        if self.use_docker_exec:  # RH MODIFIED
            cmd = " ".join(cmd) if isinstance(cmd, list) else cmd
            cmd = f"sudo docker exec {DEFAULT_DOCKER_CONTAINER_NAME} bash -c {shlex.quote(cmd)}"
        elif self.docker_user:
            cmd = " ".join(cmd) if isinstance(cmd, list) else cmd
            cmd = f"conda deactivate && {cmd}"

        command_str = self._get_command_to_run(
            cmd,
            process_stream,
            separate_stderr,
            skip_lines=skip_lines,
            source_bashrc=source_bashrc,
        )
        command = base_ssh_command + [shlex.quote(command_str)]

        log_dir = os.path.expanduser(os.path.dirname(log_path))
        os.makedirs(log_dir, exist_ok=True)

        executable = None
        if not process_stream:
            if stream_logs:
                command += [
                    f"| tee {log_path}",
                    # This also requires the executor to be '/bin/bash' instead
                    # of the default '/bin/sh'.
                    "; exit ${PIPESTATUS[0]}",
                ]
            else:
                command += [f"> {log_path}"]
            executable = "/bin/bash"

        # RH MODIFIED: Return command instead of running it
        logger.debug(f"Running command: {' '.join(command)}")
        if return_cmd:
            return " ".join(command)

        return log_lib.run_with_log(
            " ".join(command),
            log_path,
            require_outputs=require_outputs,
            stream_logs=stream_logs,
            process_stream=process_stream,
            shell=True,
            executable=executable,
            **kwargs,
        )

    def rsync(
        self,
        source: str,
        target: str,
        *,
        up: bool,
        # Advanced options.
        filter_options: Optional[str] = None,  # RH MODIFIED
        log_path: str = os.devnull,
        stream_logs: bool = True,
        max_retry: int = 1,
        return_cmd: bool = False,  # RH MODIFIED
        ignore_existing: bool = False,  # RH MODIFIED
    ) -> None:
        """Uses 'rsync' to sync 'source' to 'target'.

        Args:
            source: The source path.
            target: The target path.
            up: The direction of the sync, True for local to cluster, False
              for cluster to local.
            filter_options: The filter options for rsync.
            log_path: Redirect stdout/stderr to the log_path.
            stream_logs: Stream logs to the stdout/stderr.
            max_retry: The maximum number of retries for the rsync command.
              This value should be non-negative.
            return_cmd: If True, return the command string instead of running it.

        Raises:
            exceptions.CommandError: rsync command failed.
        """
        if self._docker_ssh_proxy_command is not None:
            docker_ssh_proxy_command = self._docker_ssh_proxy_command(["ssh"])
        else:
            docker_ssh_proxy_command = None
        ssh_options = " ".join(
            ssh_options_list(
                self.ssh_private_key,
                self.ssh_control_name,
                ssh_proxy_command=self._ssh_proxy_command,
                docker_ssh_proxy_command=docker_ssh_proxy_command,
                port=self.port,
                disable_control_master=self.disable_control_master,
            )
        )
        rsh_option = f"ssh {ssh_options}"
        return self._rsync(
            source,
            target,
            node_destination=f"{self.ssh_user}@{self.ip}",
            up=up,
            rsh_option=rsh_option,
            log_path=log_path,
            stream_logs=stream_logs,
            max_retry=max_retry,
            filter_options=filter_options,
            return_cmd=return_cmd,
            ignore_existing=ignore_existing,
        )


class SkyKubernetesRunner(KubernetesCommandRunner):
    def __init__(
        self,
        node: Tuple[str, str],
        docker_user: Optional[str] = None,
        **kwargs,
    ):
        del kwargs
        super().__init__(node)
        self.namespace, self.pod_name = node
        self.docker_user = docker_user

    def run(
        self,
        cmd: Union[str, List[str]],
        *,
        port_forward: Optional[List[int]] = None,
        require_outputs: bool = False,
        # Advanced options.
        log_path: str = os.devnull,
        # If False, do not redirect stdout/stderr to optimize performance.
        process_stream: bool = True,
        stream_logs: bool = True,
        ssh_mode: SshMode = SshMode.NON_INTERACTIVE,
        separate_stderr: bool = False,
        connect_timeout: Optional[int] = None,
        source_bashrc: bool = True,  # RH MODIFIED
        skip_lines: int = 0,
        return_cmd: bool = False,  # RH MODIFIED
        quiet_ssh: bool = None,  # RH MODIFIED
        **kwargs,
    ) -> Union[int, Tuple[int, str, str]]:
        """Uses 'kubectl exec' to run 'cmd' on a pod by its name and namespace.
        Args:
            cmd: The command to run.
            port_forward: This should be None for k8s.
            Advanced options:
            require_outputs: Whether to return the stdout/stderr of the command.
            log_path: Redirect stdout/stderr to the log_path.
            stream_logs: Stream logs to the stdout/stderr.
            check: Check the success of the command.
            ssh_mode: The mode to use for ssh.
                See SSHMode for more details.
            separate_stderr: Whether to separate stderr from stdout.
            connect_timeout: timeout in seconds for the pod connection.
            source_bashrc: Whether to source the bashrc before running the
                command.
            skip_lines: The number of lines to skip at the beginning of the
                output. This is used when the output is not processed by
                SkyPilot but we still want to get rid of some warning messages,
                such as SSH warnings.
        Returns:
            returncode
            or
            A tuple of (returncode, stdout, stderr).
        """
        # TODO(zhwu): implement port_forward for k8s.
        assert port_forward is None, (
            "port_forward is not supported for k8s " f"for now, but got: {port_forward}"
        )
        if connect_timeout is None:
            connect_timeout = _DEFAULT_CONNECT_TIMEOUT
        kubectl_args = [
            "--pod-running-timeout",
            f"{connect_timeout}s",
            "-n",
            self.namespace,
            self.pod_name,
        ]
        if ssh_mode == SshMode.LOGIN:
            assert isinstance(cmd, list), "cmd must be a list for login mode."
            base_cmd = ["kubectl", "exec", "-it", *kubectl_args, "--"]
            command = base_cmd + cmd
            proc = subprocess_utils.run(command, shell=False, check=False)
            return proc.returncode, "", ""

        kubectl_base_command = ["kubectl", "exec"]

        if ssh_mode == SshMode.INTERACTIVE:
            kubectl_base_command.append("-i")
        kubectl_base_command += [*kubectl_args, "--"]

        if self.docker_user:  # RH MODIFIED
            cmd = " ".join(cmd) if isinstance(cmd, list) else cmd
            cmd = f"conda deactivate && {cmd}"

        command_str = self._get_command_to_run(
            cmd,
            process_stream,
            separate_stderr,
            skip_lines=skip_lines,
            source_bashrc=source_bashrc,
        )
        command = kubectl_base_command + [
            # It is important to use /bin/bash -c here to make sure we quote the
            # command to be run properly. Otherwise, directly appending commands
            # after '--' will not work for some commands, such as '&&', '>' etc.
            "/bin/bash",
            "-c",
            shlex.quote(command_str),
        ]

        log_dir = os.path.expanduser(os.path.dirname(log_path))
        os.makedirs(log_dir, exist_ok=True)

        executable = None
        if not process_stream:
            if stream_logs:
                command += [
                    f"| tee {log_path}",
                    # This also requires the executor to be '/bin/bash' instead
                    # of the default '/bin/sh'.
                    "; exit ${PIPESTATUS[0]}",
                ]
            else:
                command += [f"> {log_path}"]
            executable = "/bin/bash"

        # RH MODIFIED: Return command instead of running it
        if return_cmd:
            return " ".join(command)

        return log_lib.run_with_log(
            " ".join(command),
            log_path,
            require_outputs=require_outputs,
            stream_logs=stream_logs,
            process_stream=process_stream,
            shell=True,
            executable=executable,
            **kwargs,
        )

    # @timeline.event
    def rsync(
        self,
        source: str,
        target: str,
        *,
        up: bool,
        # Advanced options.
        log_path: str = os.devnull,
        stream_logs: bool = True,
        max_retry: int = 1,
        filter_options: bool = False,  # RH MODIFIED
        return_cmd: bool = False,  # RH MODIFIED
        ignore_existing: bool = False,  # RH MODIFIED
    ) -> None:
        """Uses 'rsync' to sync 'source' to 'target'.
        Args:
            source: The source path.
            target: The target path.
            up: The direction of the sync, True for local to cluster, False
              for cluster to local.
            log_path: Redirect stdout/stderr to the log_path.
            stream_logs: Stream logs to the stdout/stderr.
            max_retry: The maximum number of retries for the rsync command.
              This value should be non-negative.
        Raises:
            exceptions.CommandError: rsync command failed.
        """

        def get_remote_home_dir() -> str:
            # Use `echo ~` to get the remote home directory, instead of pwd or
            # echo $HOME, because pwd can be `/` when the remote user is root
            # and $HOME is not always set.
            rc, remote_home_dir, stderr = self.run(
                "echo ~", require_outputs=True, separate_stderr=True, stream_logs=False
            )
            if rc != 0:
                raise ValueError(
                    "Failed to get remote home directory: "
                    f"{remote_home_dir + stderr}"
                )
            remote_home_dir = remote_home_dir.strip()
            return remote_home_dir

        # Build command.
        helper_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "kubernetes", "rsync_helper.sh"
        )
        self._rsync(
            source,
            target,
            node_destination=f"{self.pod_name}@{self.namespace}",
            up=up,
            rsh_option=helper_path,
            log_path=log_path,
            stream_logs=stream_logs,
            max_retry=max_retry,
            prefix_command=f"chmod +x {helper_path} && ",
            # rsync with `kubectl` as the rsh command will cause ~/xx parsed as
            # /~/xx, so we need to replace ~ with the remote home directory. We
            # only need to do this when ~ is at the beginning of the path.
            get_remote_home_dir=get_remote_home_dir,
            filter_options=filter_options,
            return_cmd=return_cmd,
            ignore_existing=ignore_existing,
        )



import argparse
import asyncio
import inspect
import json
import traceback
import uuid
from functools import wraps
from pathlib import Path
from typing import Dict, Optional

try:
    import ray
    import requests
    import yaml
    from fastapi import Body, FastAPI, HTTPException, Request
    from fastapi.encoders import jsonable_encoder
    from fastapi.openapi.docs import get_redoc_html, get_swagger_ui_html
    from fastapi.responses import StreamingResponse
except ImportError:
    pass

from runhouse.constants import (
    DEFAULT_HTTP_PORT,
    DEFAULT_HTTPS_PORT,
    DEFAULT_PROCESS_NAME,
    DEFAULT_SERVER_HOST,
    DEFAULT_SERVER_PORT,
)
from runhouse.globals import configs, obj_store, rns_client
from runhouse.logger import get_logger
from runhouse.resources.packages import Package
from runhouse.rns.utils.api import resolve_absolute_path, ResourceAccess
from runhouse.servers.caddy.config import CaddyConfig
from runhouse.servers.http.auth import averify_cluster_access
from runhouse.servers.http.certs import TLSCertConfig
from runhouse.servers.http.http_utils import (
    CallParams,
    CreateProcessParams,
    DeleteObjectParams,
    deserialize_data,
    folder_exists,
    folder_get,
    folder_ls,
    folder_mkdir,
    folder_mv,
    folder_put,
    folder_rm,
    FolderGetParams,
    FolderLsParams,
    FolderMvParams,
    FolderParams,
    FolderPutParams,
    FolderRmParams,
    get_token_from_request,
    handle_exception_response,
    InstallPackageParams,
    KillProcessParams,
    LogsParams,
    OutputType,
    PutObjectParams,
    PutResourceParams,
    RenameObjectParams,
    resolve_folder_path,
    Response,
    RunBashParams,
    serialize_data,
    ServerSettings,
    SetEnvVarsParams,
)
from runhouse.servers.obj_store import (
    ClusterServletSetupOption,
    ObjStoreError,
    RaySetupOption,
)
from runhouse.utils import generate_default_name, sync_function

app = FastAPI(docs_url=None, redoc_url=None)

logger = get_logger(__name__)

# TODO: Better way to store this than a global here?
running_futures: Dict[str, asyncio.Task] = {}


def validate_cluster_access(func):
    """If using Den auth, validate the user's cluster subtoken and access to the cluster before continuing."""

    @wraps(func)
    async def wrapper(*args, **kwargs):
        request: Request = kwargs.get("request")
        den_auth_enabled: bool = await HTTPServer.get_den_auth()
        is_coro = inspect.iscoroutinefunction(func)

        func_call: bool = func.__name__ in ["post_call", "get_call", "get_logs"]

        # restrict access for folder specific APIs
        access_level_required = (
            ResourceAccess.WRITE if func.__name__.startswith("folder") else None
        )
        token = get_token_from_request(request)

        request_id = request.headers.get("X-Request-ID", str(uuid.uuid4()))
        ctx_token = obj_store.set_ctx(request_id=request_id, token=token)

        try:
            if den_auth_enabled and not func_call:
                if token is None:
                    raise HTTPException(
                        status_code=401,
                        detail="No Runhouse token provided. Try running `$ runhouse login` or visiting "
                        "https://run.house/login to retrieve a token. If calling via HTTP, please "
                        "provide a valid token in the Authorization header.",
                    )
                cluster_uri = (await obj_store.aget_cluster_config()).get("name")
                cluster_access = await averify_cluster_access(
                    cluster_uri, token, access_level_required
                )
                if not cluster_access:
                    # Must have cluster access for all the non func calls
                    # Note: for func calls we handle the auth in the object store
                    raise HTTPException(
                        status_code=403,
                        detail="Cluster access is required for this operation.",
                    )

            if is_coro:
                res = await func(*args, **kwargs)
            else:
                res = func(*args, **kwargs)
        except Exception as e:
            raise e
        finally:
            if ctx_token:
                obj_store.unset_ctx(ctx_token)
        return res

    return wrapper


class HTTPServer:
    SKY_YAML = str(Path("~/.sky/sky_ray.yml").expanduser())
    memory_exporter = None

    @classmethod
    async def ainitialize(
        cls,
        conda_env_name: str = None,
        from_test: bool = False,
        *args,
        **kwargs,
    ):
        runtime_env = {"conda": conda_env_name} if conda_env_name else None

        if not configs.observability_enabled:
            logger.info("disabling cluster observability")

        # Ray and ClusterServlet should already be
        # initialized by the start script (see below)
        # But if the HTTPServer was started standalone in a test,
        # We still want to make sure the cluster servlet is initialized
        if from_test:
            await obj_store.ainitialize(
                DEFAULT_PROCESS_NAME,
                setup_ray=RaySetupOption.TEST_PROCESS,
                create_process_params=CreateProcessParams(
                    name=DEFAULT_PROCESS_NAME, runtime_env=runtime_env
                ),
            )

        # We initialize a default env servlet where some things may run.
        _ = obj_store.get_servlet(
            name=DEFAULT_PROCESS_NAME,
            create_process_params=CreateProcessParams(
                name=DEFAULT_PROCESS_NAME,
                runtime_env=runtime_env,
            ),
            create=True,
        )

    @classmethod
    def initialize(
        cls,
        conda_env_name: str = None,
        from_test: bool = False,
        *args,
        **kwargs,
    ):
        return sync_function(cls.ainitialize)(
            conda_env_name,
            from_test,
            *args,
            **kwargs,
        )

    ################################################################################################
    # Methods to expose Swagger UI and OpenAPI docs for given modules on the server
    ################################################################################################

    @staticmethod
    @app.get("/{key}/openapi.json")
    @validate_cluster_access
    async def get_openapi_spec(request: Request, key: str):
        try:
            module_openapi_spec = await obj_store.acall(
                key, method_name="openapi_spec", serialization=None
            )
        except (AttributeError, ObjStoreError):
            # The object put on the server is not an `rh.Module`, so it doesn't have an openapi_spec method
            # OR
            # The object is not found in the object store at all
            module_openapi_spec = None

        if not module_openapi_spec:
            raise HTTPException(status_code=404, detail=f"Module {key} not found.")

        return module_openapi_spec

    @staticmethod
    @app.get("/{key}/redoc")
    @validate_cluster_access
    async def get_redoc(request: Request, key: str):
        return get_redoc_html(
            openapi_url=f"/{key}/openapi.json", title="Developer Documentation"
        )

    @staticmethod
    @app.get("/{key}/docs")
    @validate_cluster_access
    async def get_swagger_ui_html(request: Request, key: str):
        return get_swagger_ui_html(
            openapi_url=f"/{key}/openapi.json",
            title=f"{key} - Swagger UI",
            oauth2_redirect_url=app.swagger_ui_oauth2_redirect_url,
            init_oauth=app.swagger_ui_init_oauth,
        )

    ################################################################################################
    # Methods to deal with server authentication logic
    ################################################################################################

    @classmethod
    async def get_den_auth(cls):
        return obj_store.is_den_auth_enabled()

    @classmethod
    async def aenable_den_auth(cls, flush: Optional[bool] = True):
        await obj_store.aenable_den_auth()

        if flush:
            await obj_store.aclear_auth_cache()

    def enable_den_auth(flush: Optional[bool] = True):
        return sync_function(HTTPServer.aenable_den_auth)(flush)

    @classmethod
    async def adisable_den_auth(cls):
        await obj_store.adisable_den_auth()

    @classmethod
    def disable_den_auth(cls):
        return sync_function(HTTPServer.adisable_den_auth)()

    @staticmethod
    @app.get("/cert")
    def get_cert():
        """Download the certificate file for this server necessary for enabling HTTPS.
        User must have access to the cluster in order to download the certificate."""

        # Default for this endpoint is "pickle" serialization
        serialization = "pickle"

        try:
            certs_config = TLSCertConfig()
            cert_path = certs_config.cert_path

            if not Path(cert_path).exists():
                raise FileNotFoundError(
                    f"No certificate found on cluster in path: {cert_path}"
                )

            with open(cert_path, "rb") as cert_file:
                cert = cert_file.read()

            return Response(
                data=serialize_data(cert, serialization),
                output_type=OutputType.RESULT_SERIALIZED,
                serialization=serialization,
            )

        except Exception as e:
            logger.exception(e)
            exception_data = {
                "error": serialize_data(e, serialization),
                "traceback": traceback.format_exc(),
            }
            return Response(
                output_type=OutputType.EXCEPTION,
                data=exception_data,
                serialization=serialization,
            )

    ################################################################################################
    # Generic utility methods for global modifications of the server
    ################################################################################################

    @staticmethod
    @app.post("/settings")
    @validate_cluster_access
    async def update_settings(request: Request, message: ServerSettings) -> Response:
        if message.cluster_name:
            await obj_store.aset_cluster_config_value("name", message.cluster_name)

        if message.den_auth:
            await HTTPServer.aenable_den_auth(flush=message.flush_auth_cache)
        elif message.den_auth is not None and not message.den_auth:
            await HTTPServer.adisable_den_auth()

        if message.autostop_mins:
            obj_store.set_cluster_config_value("autostop_mins", message.autostop_mins)

        if message.status_check_interval:
            obj_store.set_cluster_config_value(
                "status_check_interval", message.status_check_interval
            )

        return Response(output_type=OutputType.SUCCESS)

    @staticmethod
    @app.post("/install_package")
    @validate_cluster_access
    async def install_package(request: Request, params: InstallPackageParams):
        try:
            package_obj = Package.from_config(params.package_config)
            await obj_store.ainstall_package_in_all_nodes_and_processes(
                package_obj,
                conda_env_name=params.conda_env_name,
                force_sync_local=params.force_sync_local,
            )
            return Response(output_type=OutputType.SUCCESS)
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/run_bash")
    @validate_cluster_access
    async def run_bash(request: Request, params: RunBashParams):
        try:
            run_cmd_result = None
            if params.node_ip_or_idx is not None and params.process is not None:
                raise ValueError("Cannot specify both node and process.")

            elif params.node_ip_or_idx is None and params.process is None:
                # TODO: Logging when running on multiple nodes
                run_cmd_result = await obj_store.arun_bash_command_on_all_nodes(
                    command=params.command, require_outputs=params.require_outputs
                )

            elif params.node_ip_or_idx is not None:
                internal_ips = obj_store.get_internal_ips()

                if isinstance(params.node_ip_or_idx, int):
                    if params.node_ip_or_idx >= len(internal_ips):
                        raise ValueError(
                            f"Node index {params.node_ip_or_idx} out of range for cluster with {len(internal_ips)} nodes."
                        )
                    params.node_ip_or_idx = internal_ips[params.node_ip_or_idx]

                if (
                    params.node_ip_or_idx not in internal_ips
                    and params.node_ip_or_idx != "localhost"
                ):
                    raise ValueError(
                        f"Node {params.node_ip_or_idx} not a valid internal IP on the cluster."
                    )

                run_cmd_result = await obj_store.arun_bash_command_on_node_or_process(
                    command=params.command,
                    require_outputs=params.require_outputs,
                    run_name=params.run_name,
                    node_ip=params.node_ip_or_idx,
                )

            elif params.process:
                if (
                    params.process
                    not in await obj_store.aget_all_initialized_servlet_args()
                ):
                    raise ValueError(
                        f"Process {params.process} not found on the cluster."
                    )

                run_cmd_result = await obj_store.arun_bash_command_on_node_or_process(
                    command=params.command,
                    require_outputs=params.require_outputs,
                    run_name=params.run_name,
                    process=params.process,
                )

            return Response(
                output_type=OutputType.RESULT_SERIALIZED, data=run_cmd_result
            )
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    ################################################################################################
    # Logic to interact with individual processes running on the cluster
    ################################################################################################

    @staticmethod
    @app.get("/processes")
    @validate_cluster_access
    async def get_processes(request: Request):
        try:
            processes = await obj_store.alist_processes()
            return Response(
                output_type=OutputType.RESULT_SERIALIZED,
                data=serialize_data(processes, "json"),
                serialization="json",
            )
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/create_process")
    @validate_cluster_access
    async def create_process(request: Request, params: CreateProcessParams):
        try:
            _ = obj_store.get_servlet(
                name=params.name,
                create_process_params=params,
                create=True,
            )
            return Response(output_type=OutputType.SUCCESS)
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/kill_process")
    @validate_cluster_access
    async def kill_process(request: Request, params: KillProcessParams):
        try:
            await obj_store.adelete_servlet_contents(params.process_name)
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/env_vars")
    @validate_cluster_access
    async def set_env_vars(request: Request, params: SetEnvVarsParams):
        try:
            if params.process_name is not None:
                await obj_store.aset_process_env_vars(
                    params.process_name, params.env_vars
                )
            else:
                await obj_store.aset_env_vars_globally(params.env_vars)

            return Response(output_type=OutputType.SUCCESS)
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    ################################################################################################
    # Logic to interact with the filesystem
    ################################################################################################

    @staticmethod
    @app.post("/folder/method/ls")
    @validate_cluster_access
    async def folder_ls_cmd(request: Request, ls_params: FolderLsParams):
        try:
            path = resolve_folder_path(ls_params.path)
            return folder_ls(path, full_paths=ls_params.full_paths, sort=ls_params.sort)

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/mkdir")
    @validate_cluster_access
    async def folder_mkdir_cmd(request: Request, folder_params: FolderParams):
        try:
            path = resolve_folder_path(folder_params.path)
            return folder_mkdir(path)

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/get")
    @validate_cluster_access
    async def folder_get_cmd(request: Request, get_params: FolderGetParams):
        try:
            path = resolve_folder_path(get_params.path)
            return folder_get(path, mode=get_params.mode, encoding=get_params.encoding)

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/put")
    @validate_cluster_access
    async def folder_put_cmd(request: Request, put_params: FolderPutParams):
        try:
            path = resolve_folder_path(put_params.path)
            serialization = put_params.serialization
            serialized_contents = put_params.contents
            contents = deserialize_data(serialized_contents, serialization)

            return folder_put(
                path,
                contents=contents,
                overwrite=put_params.overwrite,
                mode=put_params.mode,
                serialization=serialization,
            )

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/rm")
    @validate_cluster_access
    async def folder_rm_cmd(request: Request, rm_params: FolderRmParams):
        try:
            path = resolve_folder_path(rm_params.path)
            return folder_rm(
                path, contents=rm_params.contents, recursive=rm_params.recursive
            )

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/mv")
    @validate_cluster_access
    async def folder_mv_cmd(request: Request, mv_params: FolderMvParams):
        try:
            path = resolve_folder_path(mv_params.path)
            return folder_mv(
                src_path=path,
                dest_path=mv_params.dest_path,
                overwrite=mv_params.overwrite,
            )

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/folder/method/exists")
    @validate_cluster_access
    async def folder_exists_cmd(request: Request, folder_params: FolderParams):
        try:
            path = resolve_folder_path(folder_params.path)
            return folder_exists(path=path)

        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    ################################################################################################
    # Critical object store methods for interacting with Python objects living on the cluster
    ################################################################################################

    @staticmethod
    @app.post("/resource")
    @validate_cluster_access
    async def put_resource(request: Request, params: PutResourceParams):
        try:
            process = params.process_name
            return await obj_store.aput_resource(
                serialized_data=params.serialized_data,
                serialization=params.serialization,
                process=process,
            )
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/object")
    @validate_cluster_access
    async def put_object(request: Request, params: PutObjectParams):
        try:
            await obj_store.aput(
                key=params.key,
                value=params.serialized_data,
                process=params.process_name,
                serialization=params.serialization,
                create_servlet_if_not_exists=True,
            )
            return Response(output_type=OutputType.SUCCESS)
        except Exception as e:
            return handle_exception_response(
                e,
                traceback.format_exc(),
                serialization=params.serialization,
                from_http_server=True,
            )

    @staticmethod
    @app.get("/object")
    @validate_cluster_access
    async def get_object(
        request: Request,
        key: str,
        serialization: Optional[str] = "json",
        remote: bool = False,
    ):
        try:
            return await obj_store.aget(
                key=key,
                serialization=serialization,
                remote=remote,
            )
        except Exception as e:
            return handle_exception_response(
                e,
                traceback.format_exc(),
                serialization=serialization,
                from_http_server=True,
            )

    @staticmethod
    @app.post("/rename")
    @validate_cluster_access
    async def rename_object(request: Request, params: RenameObjectParams):
        try:
            await obj_store.arename(
                old_key=params.key,
                new_key=params.new_key,
            )
            return Response(output_type=OutputType.SUCCESS)
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.post("/delete_object")
    @validate_cluster_access
    async def delete_obj(request: Request, params: DeleteObjectParams):
        try:
            if len(params.keys) == 0:
                cleared = await obj_store.akeys()
                await obj_store.aclear()
            else:
                cleared = []
                for key in params.keys:
                    await obj_store.adelete(key)
                    cleared.append(key)

            # Expicitly tell the client not to attempt to deserialize the output
            return Response(
                data=cleared,
                output_type=OutputType.RESULT_SERIALIZED,
                serialization=None,
            )
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    @app.get("/keys")
    @validate_cluster_access
    async def get_keys(request: Request, process_name: Optional[str] = None):
        try:
            if not process_name:
                output = await obj_store.akeys()
            else:
                output = await obj_store.akeys_for_servlet_name(process_name)

            # Expicitly tell the client not to attempt to deserialize the output
            return Response(
                data=output,
                output_type=OutputType.RESULT_SERIALIZED,
                serialization=None,
            )
        except Exception as e:
            return handle_exception_response(
                e, traceback.format_exc(), from_http_server=True
            )

    @staticmethod
    async def _call(
        key: str,
        method_name: Optional[str] = None,
        params: CallParams = Body(default=None),
    ):
        # Allow an empty body
        params = params or CallParams()

        try:
            if not params.run_name:
                raise ValueError("run_name is required for all calls.")
            # Call async so we can loop to collect logs until the result is ready

            fut = asyncio.create_task(
                obj_store.acall(
                    key=key,
                    method_name=method_name,
                    data=params.data,
                    stream_logs=params.stream_logs,
                    serialization=params.serialization,
                    run_name=params.run_name,
                    remote=params.remote,
                )
            )
            if params.stream_logs:
                # We'll store this future in a dictionary and just call result on it. The dictionary
                # is so the logs functionality can check if it's done and stream logs. The logs function
                # will be responsible for removing the future from memory so we don't store them indefinitely.
                running_futures[params.run_name] = fut

            return await fut
        except Exception as e:
            return handle_exception_response(
                e,
                traceback.format_exc(),
                serialization=params.serialization,
                from_http_server=True,
            )

    # TODO match "/{key}/{method_name}/{path:more_path}" for asgi / proxy requests
    @staticmethod
    @app.post("/{key}/{method_name}")
    @validate_cluster_access
    async def post_call(
        request: Request,
        key: str,
        method_name: str = None,
        params: CallParams = Body(default=None),
    ):
        return await HTTPServer._call(key, method_name, params)

    @staticmethod
    @app.get("/{key}/{method_name}")
    @validate_cluster_access
    async def get_call(
        request: Request,
        key: str,
        method_name: Optional[str] = None,
        serialization: Optional[str] = None,
        run_name: Optional[str] = None,
        stream_logs: Optional[bool] = False,
        save: Optional[bool] = False,
        remote: Optional[bool] = False,
    ):
        # Default argument to json doesn't allow a user to pass in a serialization string if they want
        # But, if they didn't pass anything, we want it to be `json` by default.
        serialization = serialization or "json"
        try:
            if run_name is None and stream_logs:
                raise ValueError(
                    "run_name is required for all calls when stream_logs is True."
                )

            if run_name is None:
                run_name = generate_default_name(
                    prefix=key if method_name == "__call__" else f"{key}_{method_name}",
                    precision="ms",  # Higher precision because we see collisions within the same second
                    sep="--",
                )

            # The types need to be explicitly specified as parameters first so that
            # we can cast Query params to the right type.
            params = CallParams(
                serialization=serialization,
                run_name=run_name,
                stream_logs=stream_logs,
                save=save,
                remote=remote,
            )

            query_params_remaining = dict(request.query_params)
            call_params_dict = params.model_dump()
            for k, v in dict(request.query_params).items():
                # If one of the query_params matches an arg in CallParams, set it
                # And also remove it from the query_params dict, so the rest
                # of the args will be passed as kwargs
                if k in call_params_dict:
                    del query_params_remaining[k]

            data = {
                "args": [],
                "kwargs": query_params_remaining,
            }
            params.data = serialize_data(data, serialization)

            return await HTTPServer._call(key, method_name, params)
        except Exception as e:
            logger.exception(e)
            exception_data = {
                "error": serialize_data(e, serialization),
                "traceback": traceback.format_exc(),
            }
            return Response(
                output_type=OutputType.EXCEPTION,
                data=exception_data,
                serialization=serialization,
            )

    # `/logs` POST endpoint that takes in request and LogParams
    @staticmethod
    @app.post("/logs")
    @validate_cluster_access
    async def get_logs(
        request: Request,
        params: LogsParams = Body(default=None),
    ):
        try:
            if (
                sum(
                    arg is not None
                    for arg in [params.key, params.node_ip_or_idx, params.process]
                )
                != 1
            ):
                raise ValueError(
                    "Exactly one of key, node_ip, or process must be provided to get logs"
                )

            if isinstance(params.node_ip_or_idx, int):
                internal_ips = obj_store.get_internal_ips()
                if params.node_ip_or_idx >= len(internal_ips):
                    raise ValueError(
                        f"Node index {params.node_ip_or_idx} out of range for cluster with {len(internal_ips)} nodes."
                    )
                params.node_ip_or_idx = internal_ips[params.node_ip_or_idx]

            return StreamingResponse(
                HTTPServer._get_logs_generator(
                    run_name=params.run_name,
                    key=params.key,
                    node_ip=params.node_ip_or_idx,
                    process=params.process,
                    serialization=params.serialization,
                ),
                media_type="application/json",
            )
        except Exception as e:
            logger.exception(e)
            exception_data = {
                "error": serialize_data(e, params.serialization),
                "traceback": traceback.format_exc(),
            }
            return Response(
                output_type=OutputType.EXCEPTION,
                data=exception_data,
                serialization=params.serialization,
            )

    @staticmethod
    async def _get_logs_generator(
        run_name: str,
        key: Optional[str] = None,
        node_ip: Optional[str] = None,
        process: Optional[str] = None,
        serialization=None,
    ):
        logger.info(f"Streaming logs for run name: '{run_name}'")
        try:
            async for log_lines in obj_store.alogs_for_run_name(
                run_name=run_name,
                key=key,
                node_ip=node_ip,
                servlet_name=process,
            ):
                logger.debug(f"Yielding logs for key {run_name}")
                yield json.dumps(
                    jsonable_encoder(
                        Response(
                            data=log_lines,
                            output_type=OutputType.STDOUT,
                        )
                    )
                ) + "\n"
        except Exception as e:
            logger.exception(e)
            # NOTE: We do not convert the exception to an HTTPException here, because once we're inside this
            # generator starlette has already returned the StreamingResponse and there is no way to halt the stream
            # to return a 403 instead. Users shouldn't notice much of a difference, because if working through the
            # client the exception will be serialized and returned to appear as a native python exception, and if
            # working through an HTTP call stream_logs is False by default, so a normal HTTPException will be raised
            # above before entering this generator.
            exception_data = {
                "error": serialize_data(e, serialization),
                "traceback": traceback.format_exc(),
            }
            yield json.dumps(
                jsonable_encoder(
                    Response(
                        output_type=OutputType.EXCEPTION,
                        data=exception_data,
                        serialization=serialization,
                    )
                )
            )

    ################################################################################################
    # Cluster status and metadata methods
    ################################################################################################

    @staticmethod
    @app.get("/check")
    def check_server():

        # Default for this endpoint is "pickle" serialization
        serialization = "pickle"

        try:
            if not ray.is_initialized():
                raise Exception("Ray is not initialized, restart the server.")
            logger.info("Server is up.")

            import runhouse

            return {"rh_version": runhouse.__version__}
        except Exception as e:
            logger.exception(e)
            exception_data = {
                "error": serialize_data(e, serialization),
                "traceback": traceback.format_exc(),
            }
            return Response(
                output_type=OutputType.EXCEPTION,
                data=exception_data,
                serialization=serialization,
            )

    @staticmethod
    @app.get("/status")
    @validate_cluster_access
    def get_status(request: Request, send_to_den: bool = False):

        return obj_store.status(send_to_den=send_to_den)

    @staticmethod
    def _collect_cluster_stats():
        """Collect cluster metadata and send to Grafana Loki"""

        cluster_data = HTTPServer._cluster_status_report()
        sky_data = HTTPServer._cluster_sky_report()

        HTTPServer._log_cluster_data(
            {**cluster_data, **sky_data},
            labels={"username": configs.username, "environment": "prod"},
        )

    @staticmethod
    def _cluster_status_report():
        import ray._private.usage.usage_lib as ray_usage_lib
        from ray._raylet import GcsClient

        gcs_client = GcsClient(address="127.0.0.1:6379", nums_reconnect_retry=20)

        # fields : ['ray_version', 'python_version']
        cluster_metadata = ray_usage_lib.get_cluster_metadata(gcs_client)

        # fields: ['total_num_cpus', 'total_num_gpus', 'total_memory_gb', 'total_object_store_memory_gb']
        cluster_status_report = ray_usage_lib.get_cluster_status_to_report(
            gcs_client
        ).__dict__

        return {**cluster_metadata, **cluster_status_report}

    @staticmethod
    def _cluster_sky_report():
        try:
            with open(HTTPServer.SKY_YAML, "r") as stream:
                sky_ray_data = yaml.safe_load(stream)
        except FileNotFoundError:
            # For non on-demand clusters we won't have sky data
            return {}

        provider = sky_ray_data["provider"]
        node_config = sky_ray_data["available_node_types"].get("ray.head.default", {})

        return {
            "cluster_name": sky_ray_data.get("cluster_name"),
            "region": provider.get("region"),
            "provider": provider.get("module"),
            "instance_type": node_config.get("node_config", {}).get("InstanceType"),
        }

    @staticmethod
    def _log_cluster_data(data: dict, labels: dict):
        from runhouse.rns.utils.api import log_timestamp

        payload = {
            "streams": [
                {"stream": labels, "values": [[str(log_timestamp()), json.dumps(data)]]}
            ]
        }

        payload = json.dumps(payload)
        resp = requests.post(
            f"{rns_client.api_server_url}/admin/logs", data=json.dumps(payload)
        )

        if resp.status_code == 405:
            # api server not configured to receive grafana logs
            return

        if resp.status_code != 200:
            logger.error(
                f"({resp.status_code}) Failed to send logs to Grafana Loki: {resp.text}"
            )


async def main():
    import uvicorn

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--host",
        type=str,
        default=None,
        help=f"Host to run server on. By default will run on {DEFAULT_SERVER_HOST}",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=None,
        help="Port to run daemon on on. If provided and Caddy is not enabled, "
        f"will attempt to run the daemon on this port, defaults to {DEFAULT_SERVER_PORT}",
    )
    parser.add_argument(
        "--conda-env", type=str, default=None, help="Conda env to run server in"
    )
    parser.add_argument(
        "--use-https",
        action="store_true",  # if providing --use-https will be set to True
        default=argparse.SUPPRESS,  # If user didn't specify, attribute will not be present (not False)
        help="Start an HTTPS server with new TLS certs",
    )
    parser.add_argument(
        "--use-den-auth",
        action="store_true",  # if providing --use-den-auth will be set to True
        default=argparse.SUPPRESS,  # If user didn't specify, attribute will not be present (not False)
        help="Whether to authenticate requests with a Runhouse token",
    )
    parser.add_argument(
        "--domain",
        type=str,
        default=None,
        help="Server domain name",
    )
    parser.add_argument(
        "--ssl-keyfile",
        type=str,
        default=None,
        help="Path to SSL key file on the cluster to use for HTTPS",
    )
    parser.add_argument(
        "--ssl-certfile",
        type=str,
        default=None,
        help="Path to SSL cert file on the cluster to use for HTTPS",
    )
    parser.add_argument(
        "--restart-proxy",
        action="store_true",  # if providing --restart-proxy will be set to True
        help="Reconfigure Caddy",
    )
    parser.add_argument(
        "--use-caddy",
        action="store_true",  # if providing --use-caddy will be set to True
        default=argparse.SUPPRESS,  # If user didn't specify, attribute will not be present (not False)
        help="Configure Caddy as a reverse proxy",
    )
    parser.add_argument(
        "--certs-address",
        type=str,
        default=None,
        help="Address to use for generating self-signed certs and enabling HTTPS. (e.g. public IP address)",
    )
    parser.add_argument(
        "--api-server-url",
        type=str,
        default=rns_client.api_server_url,
        help="URL of Runhouse Den",
    )
    parser.add_argument(
        "--from-python",
        action="store_true",
        default=argparse.SUPPRESS,
        help="Whether HTTP server is called from Python rather than CLI.",
    )

    parse_args = parser.parse_args()

    conda_env_name = parse_args.conda_env
    restart_proxy = parse_args.restart_proxy
    api_server_url = parse_args.api_server_url

    # The object store and the cluster servlet within it need to be
    # initialized in order to call `obj_store.get_cluster_config()`, which
    # uses the object store to load the cluster config from Ray.
    # When setting up the server, we always want to create a new ClusterServlet.
    # We only want to forcibly start a Ray cluster if asked.
    # We connect this to the "base" process, which we'll initialize later,
    # so writes to the obj_store within the server get proxied to the "base" process.
    await obj_store.ainitialize(
        DEFAULT_PROCESS_NAME,
        setup_cluster_servlet=ClusterServletSetupOption.FORCE_CREATE,
    )

    cluster_config = await obj_store.aget_cluster_config()
    if not cluster_config:
        logger.warning(
            "Cluster config is not set. Using default values where possible."
        )
    else:
        logger.info("Loaded cluster config from Ray.")

    logger.info("Initalizing all individual node servlets.")
    await obj_store.ainitialize_node_servlets()

    ########################################
    # Handling args that could be specified in the
    # cluster_config.json or via CLI args
    ########################################

    # Den URL
    cluster_config["api_server_url"] = api_server_url

    # Server port
    if parse_args.port is not None and parse_args.port != cluster_config.get(
        "server_port"
    ):
        logger.warning(
            f"CLI provided server port: {parse_args.port} is different from the server port specified in "
            f"cluster_config.json: {cluster_config.get('server_port')}. Prioritizing CLI provided port."
        )

    port_arg = (
        parse_args.port or cluster_config.get("server_port") or DEFAULT_SERVER_PORT
    )

    if port_arg is not None:
        cluster_config["server_port"] = port_arg

    # Den auth enabled
    if hasattr(
        parse_args, "use_den_auth"
    ) and parse_args.use_den_auth != cluster_config.get("den_auth", False):
        logger.warning(
            f"CLI provided den_auth: {parse_args.use_den_auth} is different from the den_auth specified in "
            f"cluster_config.json: {cluster_config.get('den_auth')}. Prioritizing CLI provided den_auth."
        )

    den_auth = getattr(parse_args, "use_den_auth", False) or cluster_config.get(
        "den_auth", False
    )
    cluster_config["den_auth"] = den_auth

    if den_auth:
        await obj_store.aenable_den_auth()

    domain = parse_args.domain or cluster_config.get("domain", None)
    cluster_config["domain"] = domain

    # Keyfile
    if (
        parse_args.ssl_keyfile is not None
        and parse_args.ssl_keyfile != cluster_config.get("ssl_keyfile")
    ):
        logger.warning(
            f"CLI provided ssl_keyfile: {parse_args.ssl_keyfile} is different from the ssl_keyfile specified in "
            f"cluster_config.json: {cluster_config.get('ssl_keyfile')}. Prioritizing CLI provided ssl_keyfile."
        )

    ssl_keyfile_path_arg = parse_args.ssl_keyfile or cluster_config.get("ssl_keyfile")
    if ssl_keyfile_path_arg is not None:
        cluster_config["ssl_keyfile"] = ssl_keyfile_path_arg
    parsed_ssl_keyfile = (
        resolve_absolute_path(ssl_keyfile_path_arg) if ssl_keyfile_path_arg else None
    )

    # Certfile
    if (
        parse_args.ssl_certfile is not None
        and parse_args.ssl_certfile != cluster_config.get("ssl_certfile")
    ):
        logger.warning(
            f"CLI provided ssl_certfile: {parse_args.ssl_certfile} is different from the ssl_certfile specified in "
            f"cluster_config.json: {cluster_config.get('ssl_certfile')}. Prioritizing CLI provided ssl_certfile."
        )

    ssl_certfile_path_arg = parse_args.ssl_certfile or cluster_config.get(
        "ssl_certfile"
    )
    if ssl_certfile_path_arg is not None:
        cluster_config["ssl_certfile"] = ssl_certfile_path_arg
    parsed_ssl_certfile = (
        resolve_absolute_path(ssl_certfile_path_arg) if ssl_certfile_path_arg else None
    )

    # Host
    if parse_args.host is not None and parse_args.host != cluster_config.get(
        "server_host"
    ):
        logger.warning(
            f"CLI provided server_host: {parse_args.host} is different from the server_host specified in "
            f"cluster_config.json: {cluster_config.get('server_host')}. Prioritizing CLI provided server_host."
        )

    host = parse_args.host or cluster_config.get("server_host") or DEFAULT_SERVER_HOST
    cluster_config["server_host"] = host

    # Address in the case we're a TLS server
    if (
        parse_args.certs_address is not None
        and parse_args.certs_address != cluster_config.get("ips", [None])[0]
    ):
        logger.warning(
            f"CLI provided certs_address: {parse_args.certs_address} is different from the certs_address specified in "
            f"cluster_config.json: {cluster_config.get('ips', [None])[0]}. Prioritizing CLI provided certs_address."
        )

    certs_addresses = (
        parse_args.certs_address
        or cluster_config.get("ips", None)
        or cluster_config.get("compute_properties", {}).get("ips", None)
    )

    # We don't want to unset multiple addresses if they were set in the cluster config
    if certs_addresses is not None:
        cluster_config["ips"] = certs_addresses
    else:
        cluster_config["ips"] = ["0.0.0.0"]

    certs_address = certs_addresses[0] if certs_addresses else None

    config_conn = cluster_config.get("server_connection_type")

    # Use caddy as reverse proxy
    if hasattr(parse_args, "use_caddy") and parse_args.use_caddy != (
        config_conn in ["tls", "none"]
    ):
        logger.warning(
            f"CLI provided use_caddy: {parse_args.use_caddy} is different from the server_connection_type specified in "
            f"cluster_config.json: {config_conn}. Prioritizing CLI provided use_caddy."
        )

    # Use HTTPS
    if hasattr(parse_args, "use_https") and parse_args.use_https != (
        config_conn == "tls"
    ):
        logger.warning(
            f"CLI provided use_https: {parse_args.use_https} is different from the server_connection_type specified in "
            f"cluster_config.json: {config_conn}. Prioritizing CLI provided use_https."
        )

    use_caddy = getattr(parse_args, "use_caddy", False) or (
        config_conn in ["tls", "none"]
    )
    use_https = getattr(parse_args, "use_https", False) or (config_conn == "tls")
    cluster_config["server_connection_type"] = (
        "tls" if use_https else "none" if use_caddy else config_conn
    )

    # If there was no `cluster_config.json`, then server was created
    # simply with `runhouse server start`.
    # A real `cluster_config` that was loaded
    # from json would have this set for sure
    if not cluster_config.get("resource_subtype"):
        # This is needed for the Cluster object to be created
        # in rh.here.
        cluster_config["resource_subtype"] = "Cluster"

        # server_connection_type is not set up if this is done through
        # a local `runhouse server start`
        cluster_config["server_connection_type"] = "tls" if use_https else "none"

    await obj_store.aset_cluster_config(cluster_config)

    await HTTPServer.ainitialize(
        default_process_name=DEFAULT_PROCESS_NAME,
        conda_env=conda_env_name,
    )

    if den_auth:
        # Update den auth if enabled - keep as a class attribute to be referenced by the validator decorator
        await HTTPServer.aenable_den_auth()

    if use_https and not domain:
        # If using https (whether or not Caddy is being used) and no domain is specified, need to provide both
        # key and cert files - they should both already exist on the cluster
        if (
            not parsed_ssl_keyfile
            or not Path(parsed_ssl_keyfile).exists()
            or not parsed_ssl_certfile
            or not Path(parsed_ssl_certfile).exists()
        ):
            # Custom certs should already be on the cluster if their file paths are provided
            raise FileNotFoundError(
                f"Could not find SSL private key and cert files on the cluster, which are required when specifying "
                f"a port ({port_arg}) or enabling HTTPS. Please specify the paths using the --ssl-certfile and "
                f"--ssl-keyfile flags."
            )

    # If the daemon port was not specified, it should be the default RH port
    daemon_port = port_arg
    if not daemon_port or daemon_port in [
        DEFAULT_HTTP_PORT,
        DEFAULT_HTTPS_PORT,
    ]:
        # Since one of HTTP_PORT or HTTPS_PORT was specified, Caddy is set up to forward requests
        # from the daemon to that port
        daemon_port = DEFAULT_SERVER_PORT

    # Note: running the FastAPI app on a higher, non-privileged port (8000) and using Caddy as a reverse
    # proxy to forward requests from port 80 (HTTP) or 443 (HTTPS) to the app's port.
    if use_caddy:
        logger.debug("Using Caddy as a reverse proxy")
        if certs_address is None and domain is None:
            raise ValueError(
                "Must provide the server address or domain to configure Caddy. No address or domain found in the "
                "server start command (--certs-address or --domain) or in the cluster config YAML saved on the cluster."
            )

        cc = CaddyConfig(
            address=certs_address,
            domain=domain,
            rh_server_port=daemon_port,
            ssl_key_path=parsed_ssl_keyfile,
            ssl_cert_path=parsed_ssl_certfile,
            use_https=use_https,
            force_reinstall=restart_proxy,
        )
        cc.configure()

        caddy_port = DEFAULT_HTTPS_PORT if use_https else DEFAULT_HTTP_PORT
        logger.info(f"Caddy is proxying requests from {caddy_port} to {daemon_port}.")

    logger.info(
        f"Launching Runhouse API server with den_auth={den_auth} and "
        + f"on host={host} and use_https={use_https} and port_arg={daemon_port}"
    )

    cluster = None
    if not hasattr(parse_args, "from_python"):
        from runhouse.resources.hardware import Cluster

        cluster = Cluster.from_config(cluster_config)
        cluster._sync_image_to_cluster()  # TODO - image needs to be part of the cluster config

    # Only launch uvicorn with certs if HTTPS is enabled and not using Caddy
    uvicorn_cert = parsed_ssl_certfile if not use_caddy and use_https else None
    uvicorn_key = parsed_ssl_keyfile if not use_caddy and use_https else None

    config = uvicorn.Config(
        app,
        host=host,
        port=daemon_port,
        ssl_certfile=uvicorn_cert,
        ssl_keyfile=uvicorn_key,
        loop="uvloop",
    )
    server = uvicorn.Server(config)
    await server.serve()

    # TODO - image env vars (+ secrets?)


if __name__ == "__main__":
    asyncio.run(main())



import copy
import os
from pathlib import Path

from typing import Dict

from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret
from runhouse.resources.secrets.utils import _check_file_for_mismatches
from runhouse.utils import create_local_dir


class HuggingFaceSecret(ProviderSecret):
    """
    .. note::
            To create a HuggingFaceSecret, please use the factory method :func:`provider_secret` with
            ``provider="huggingface"``.
    """

    # values format: {"token": hf_token}
    _PROVIDER = "huggingface"
    _DEFAULT_CREDENTIALS_PATH = "~/.cache/huggingface/token"
    _DEFAULT_ENV_VARS = {"token": "HF_TOKEN"}

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        return HuggingFaceSecret(**config, dryrun=dryrun)

    def _write_to_file(
        self,
        path: str,
        values: Dict = None,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        new_secret = copy.deepcopy(self)
        if not _check_file_for_mismatches(
            path, self._from_path(path), values, overwrite
        ):
            token = values["token"]
            full_path = create_local_dir(path)
            with open(full_path, "a") as f:
                f.write(token)

            if write_config:
                new_secret._add_to_rh_config(path)

        new_secret._values = None
        new_secret.path = path
        return new_secret

    def _from_path(self, path: str = None):
        token = None
        if path and os.path.exists(os.path.expanduser(path)):
            token = Path(os.path.expanduser(path)).read_text().strip("\n")
        if token:
            return {"token": token}
        return {}
